(in-package :http)

;;; Configurable server parameters 

;;; These can be changed by specifying command line options in case
;;; you have built an image, or by putting new values for them in the
;;; configuration file. Priority order is: command-line, configuration
;;; file, default.

(defvar *machine-name* (machine-instance))

(defvar *http-port* 3000 
  "TCP port number for incoming HTTP connections. 
Equivalent command-line option: -hp, -http-port")

(defvar *eval-port* 3001
  "TCP port number for incoming connections to the Listener.
Equivalent command-line option: -ep, -eval-port")

(defvar *LispWeb-root-dir* (translate-logical-pathname "food:web;")
  "Root dir of the server file-system.
Equivalent command-line option: -dir")

(defvar *LispWeb-conf-dir* (translate-logical-pathname "food:code;http;conf;")
  "Directory where configuration files are stored.")

(defvar *LispWeb-init-file* "lispweb-init.lisp" 
  "File to load at server startup.
Equivalent command-line option: -i, -init")

(defvar *LispWeb-log-file* "lispweb-log"
  "Server log filename.
Equivalent command-line option: -lf, -log-file")
 
(defvar *LispWeb-debug-file* nil
  "Server debug filename.
Equivalent command-line option: -df, -debug-file")

(defvar *LispWeb-timing-file* nil
  "Server timing info filename.
Equivalent command-line option: -tf, -timing-file")

(defvar *LispWeb-visible-pages* nil
  "Whether Lisp source should be visible by default")

(defvar *LispWeb-robots-txt*
  "User-agent: *
Disallow: /
"
  "Contents of /robots.txt file for this server.")
  
(defvar *http-delay* 0
  "Number of seconds to sleep between polls.
Equivalent command-line option: -hd, -http-delay")

(defvar *http-keepalive-timeout* 10
  "Timeout for Keep-Alive connections.")

(defvar *http-multithreaded-p* t
  "Handle multiple requests in parallel?
Equivalent command-line option: -m, -multi")

(defvar *operator-address* "alb@ipvaim.unipv.it"
  "Email address of the server operator.
Equivalent command-line option: -oa, -op-add")

(defvar *operator-password* "Sultans Of Swing" 
  "Password to use the control page and the Network Listener.
Equivalent command-line option: -op, -op-pass")

;;; Reserved server parameters

(defconstant *LispWeb-version* "1.1"
  "The server version.")

(defconstant *LispWeb-platform*
    #+LUCID "lucid"
    #+ALLEGRO "acl"
    #+CLISP "clisp"
    #+LISPWORKS "lispworks"
    "The platform on which the server was compiled.")

(defconstant *LispWeb-signature* 
    (format nil "LispWeb ~a (~a)" 
	    *LispWeb-Version*
	    *LispWeb-platform*))

(defconstant *LispWeb-base-header*
    (format nil "MIME-Version: 1.0
Server: ~a
Generator: ~a~%"
	    *LispWeb-signature*
	    *LispWeb-signature*))
 
(defconstant *lispweb-lambda-keywords*
    nil) ;((map-x 0) (map-y 0)))

(defconstant *lispweb-declarations*
    ;;    #+(or ALLEGRO CLISP) '(ignorable map-x map-y)
    ;;    #+LUCID '(ignore map-x map-y)
    '(special lispweb-page-key)
    )

(defconstant *LispWeb-banner*
    "<H3>LispWeb Common Lisp Http Server</H3><HR>")
    
(defvar *LispWeb-server-hostname* nil
  "Fully-qualified domain name of the server.")

(defvar *LispWeb-build-date* nil
  "Server build date.")

(defvar *LispWeb-services* nil
  "List of services loaded in the server.")

(defvar *LispWeb-processes* nil
  "List of active server processes. Each process is identified 
by either a keyword (:http, :telnet) for servers, 
or a number for connection handlers.")

(defvar *LispWeb-log-stream* nil
  "Server log stream.")

(defvar *LispWeb-debug-stream* nil
  "Server debug stream.")

(defvar *LispWeb-timing-stream* nil
  "Server timing statistics stream.")

(defvar *LispWeb-debug* nil
  "Enter debugger in case of errors?")

(defvar *LispWeb-allow-html-files* nil
  "Allow serving regular HTML files?")

(defvar *LispWeb-html-indent* 0
  "Indentation of HTML code")

(defvar *LispWeb-hall-demo* nil)

(defvar *LispWeb-traced-pages* nil
  "List of traced pages (for debugging purposes)")

(defvar *http-socket* nil
  "TCP socket to use for HTTP output.")

(defvar *eval-socket* nil
  "TCP socket to use for output by the Listener.")

(defvar *http-stream* nil
  "Lisp stream to use for output.")

(defvar *eval-stream* nil
  "Lisp stream to use for output.")

(defvar *http-connection* nil
  "Sockaddr structure for incoming HTTP connections.")

(defvar *listener-connection* nil
  "Sockaddr structure for incoming connections to the Listener.")

(defvar *closing-server* nil
  "Set to T to shut the server down.")

;;; Connection information

(defstruct http-info
  socket				;the network socket
  stream				;the I/O stream
  fd					;the unix file descriptor
  addr					;Internet address of the client
  request-counter			;progressive number of the request
  request-line				;the request line
  request-object			;the requested object
  arguments				;the page arguments
  method				;the method name
  browser				;browser family
  protocol				;the protocol name
  headers				;header fields
  fields				;form fields
  code					;response code
  time)					;time needed to answer

(defstruct eval-info
  stream				;the I/O stream
  socket				;the network socket
  addr)					;Internet address of the client

(defvar *http-info* (make-http-info :addr "127.0.0.1")
  "An HTTP-CONNECTION-INFO structure")

(defvar *eval-info* (make-eval-info :addr "127.0.0.1")
  "An EVAL-CONNECTION-INFO structure")

;;; Header field structure

(defstruct header-field
  name
  value
  options)

;;; Server statistics

(defvar *request-counter* 0 
  "Count the number of requests received by the server since startup.")

(defvar *startup-date* nil 
  "Date and time of server startup.")

;;; Server objects

(defvar *system-pages* nil
  "Tree of page requests handled by the system.")

(defvar *user-pages* nil
  "Tree of page requests handled by the user.")

(defvar *top-level-links* nil
  "List of links that must appear in the top page.")

;;; Browser types

(defconstant *LispWeb-browser-table*
    '((" NCSA Mosaic" . :mosaic)
      ("Mozilla" . :netscape)
      ("Lynx" . :lynx)
      ("" . :generic)))


(defvar *home-user-pathname* (translate-logical-pathname "food:code;"))

;;;added by john domingue
;;;i'm gonna have to work out how to automatically get this one day
(defvar tcp::*LispWeb-server-domain-name* 
  (machine-domain) 
  ;;  "open.ac.uk" wsmo tutorial hardwire domain name
  ;;"swi.psy.uva.nl" ;;amsterdam
  "The domain name where the server sits")

(defvar tcp::*host-ip-address* 
  ;;"137.108.24.105"
  ;;"localhost"
  (machine-address) ;;wsmo tutorial hardwire domain name
  ;;for ibrow review 
  ;;"137.108.24.105"
  ;;"145.18.154.235" ;;amsterdam
  "Calcuate the ip address when the machine is unplugged.")

(defvar tcp::*unplugged-p* nil
  "For doing demos when the machine is unplugged.")

(defun cl-user::unplugged ()
  (setf tcp::*unplugged-p* t
        *http-info* (make-http-info :addr tcp::*host-ip-address*)
        *eval-info* (make-eval-info :addr tcp::*host-ip-address*)))

(defun cl-user::plugged ()
  (setf tcp::*unplugged-p* nil))

(defun cl-user::plugged-p ()
  (not tcp::*unplugged-p*))

(defun current-host-ip-address ()
  (if tcp::*unplugged-p*
      "localhost"
    (machine-ip-address)
    ;;"137.108.24.105"
    ;;"145.18.154.235"
    )) ;;amsterdam

