(in-package :http)

;;; Header codes

(defconstant *status-code-alist*
    '((:continue       100 "Continue")
      (:switching      101 "Switching Protocols")
      
      (:ok             200 "OK")
      (:created        201 "Created")
      (:accepted       202 "Accepted")
      (:provisional    203 "Provisional Information")
      (:empty          204 "No Response")
      (:deleted        205 "Deleted")
      (:modified       206 "Modified")

      (:moved-perm     301 "Document Moved Permanently")
      (:redirect       302 "Document Moved Temporarily")
      (:forward        303 "Method")
      (:not-mod        304 "Not Modified")
      (:use-proxy      305 "Use Proxy")
       
      (:bad-syntax     400 "Bad Request")
      (:no-auth        401 "Unauthorized")
      (:payment        402 "Payment Required")
      (:forbidden      403 "Forbidden")
      (:not-found      404 "Not Found")
      (:not-allowed    405 "Method Not Allowed")
      (:not-acceptable 406 "None Acceptable")
      (:proxy-auth     407 "Proxy Authentication Required")
      (:timeout        408 "Request Timeout")
      (:conflict       409 "Conflict")
      (:gone           410 "Gone")
      (:len-required   411 "Length Required")
      (:unless-true    412 "Unless True")
      
      (:error          500 "Server Error")
      (:unimplemented  501 "Not Implemented")
      (:bad-gateway    502 "Bad Gateway")
      (:unavailable    503 "Service Unavailable")
      (:gw-timeout     504 "Gateway Timeout")))

(defun code-to-status (code)
  (find-if #'(lambda (item)
	       (= (second item) code))
	   *status-code-alist*))

;;; Header classes

(defclass http-base-header ()
  ())

(defclass http-generic-header (http-base-header)
  ())

(defclass http-text-header (http-base-header)
  ())

;vlad 28/02/06
(defclass http-xml-header (http-base-header)
  ())

(defclass http-animation-header (http-base-header)
  ())

;;; Header instances

(defvar *http-base-header* (make-instance 'http-base-header))
    
(defvar *http-generic-header* (make-instance 'http-generic-header))
    
(defvar *http-text-header* (make-instance 'http-text-header))

;vlad 28/02/06
(defvar *http-xml-header* (make-instance 'http-xml-header))

(defvar *http-animation-header* (make-instance 'http-animation-header))

;;; Condition to handle HEAD requests

(define-condition head-request
    (condition)
  nil)

(defvar *head-request-condition*
    (make-condition 'head-request))

;;; Methods

(defun http-header (header &rest keys)
  (format *http-stream* "~a" 
          (with-output-to-string (header-stream)
            (apply #'http-send-header header-stream header keys)))
  (terpri *http-stream*)
  (http-force-output *http-stream*)
  (when (eq (http-info-method *http-info*) :head)
    (error *head-request-condition*)))	;exit after header for HEAD requests

(defun output-response-line (code-name)
  (let ((code (if (numberp code-name)
		  code-name
		(assoc code-name *status-code-alist*))))
    (setf (http-info-code *http-info*) code-name)
    (format *http-stream* "HTTP/1.0 ~d ~a~%" (second code) (third code))))

(defgeneric http-send-header (stream header &key &allow-other-keys)
  )
  
(defmethod http-send-header (stream (h http-base-header) &key other-headers &allow-other-keys)
  (if (listp other-headers)
      (format stream "~a~{~a~%~}" *lispweb-base-header* other-headers)
    (format stream "~a~a~%" *lispweb-base-header* other-headers))
  (when (equalp (header-value :connection) "Keep-Alive")
    (format stream "Connection: Keep-Alive~%")
    (format stream "Keep-Alive: timeout=~d~%" *http-keepalive-timeout*)))

(defmethod http-send-header :around (stream (h http-generic-header) &key type (length 0) &allow-other-keys)
  (output-response-line :ok)
  (call-next-method)
  (format stream "Date: ~a
Last-Modified: ~a
Content-Type: ~a
Content-Length: ~d~%" 
	  (http-time) 
	  (http-time) 
	  type 
	  length))

(defmethod http-send-header :around (stream (h http-text-header) &key code-name (length 0) 
				     &allow-other-keys)
  (output-response-line code-name)
  (call-next-method)
  (format stream "Date: ~a
Last-Modified: ~a
Content-Type: text/html
Content-Length: ~d~%" 
	    (http-time) 
	    (http-time)
            length))

(defmethod http-send-header :around (stream (h http-text-header) &key code-name (length 0) 
				     &allow-other-keys)
  (output-response-line code-name)
  (call-next-method)
  (format stream "Date: ~a
Last-Modified: ~a
Content-Type: text/html
Content-Length: ~d~%" 
	    (http-time) 
	    (http-time)
            length))

;vlad 28/02/06
(defmethod http-send-header :around (stream (h http-xml-header) &key code-name (length 0) 
				     &allow-other-keys)
  (output-response-line code-name)
  (call-next-method)
  (format stream "Date: ~a
Last-Modified: ~a
Content-Type: text/xml
Content-Length: ~d~%" 
	    (http-time) 
	    (http-time)
            length))

(defmethod http-send-header :around (stream (h http-animation-header) &key boundary &allow-other-keys)
  (output-response-line :ok)
  (call-next-method)
  (format stream "Date: ~a
Last-Modified: ~a
Content-Type: multipart/x-mixed-replace;boundary=~s~%" 
	  (http-time) 
	  (http-time) 
	  boundary))
