;;; Code that implements the HTTP methods.

(in-package :http)

;;; Dispatchers

(defun ordered-insert (item list test)
  "Insert item into list preserving the ordering established by test."
  
  (declare (type (or cons null) list))
  (cond ((null list) (list item))
	((funcall test item (car list))
	 (cons item list))
	((equal (car item) (caar list))
	 (unless (eq (cdr item) (cdr (car list)))
	   (warn "~a replaces ~a on url ~s" (cdr item) (cdr (car list)) (car item)))
	 (cons item (cdr list)))
	(t (cons (car list) 
		 (ordered-insert item (cdr list) test)))))

(defun add-obj (tree name definition)
  (ordered-push (cons name definition) 
		tree
		#'(lambda (x y)
		    (> (length (car x))
		       (length (car y))))))

(defun call-handler (tree request &optional arguments)
  (when tree
    (let* ((key (car (car tree)))
	   (len (length key)))
      (if (and (>= (length request) len)
	       (string= key (subseq request 0 len)))
	  (let ((fun (cdr (car tree)))
		(tail (subseq request len)))
	    (if (keywordp fun)
		(http-deliver fun tail)
	      (case (get fun :lispweb-fun-type)
		(:multiple (list* fun (append (explode-request tail) arguments)))
		(:page (progn
			 (setq arguments (append (explode-request tail) arguments))
			 (setf (http-info-arguments *http-info*) arguments)
			 (apply fun arguments)))
		(t (call-handler (symbol-value fun) tail arguments)))))
	(call-handler (cdr tree) request arguments)))))

(defun delete-handler (tree request)	;unimplemented
  )
  
;;; Applications

(defun application-path (application)
  (get application :lispweb-path))

;;; HTTP-reply
(defgeneric http-reply (method request)
  (:documentation "Reply to the received request. Can be specialized to handle additional methods."))

(defmethod http-reply (method request)
  (declare (ignore request))
  (http-header *http-text-header* :code-name :unimplemented))

(defmethod http-reply ((method (eql :get)) request)
  (when (string= request "/")
    (setq request "/top"))  
  (call-handler *lispweb-type-table* request))

(defmethod http-reply ((method (eql :post)) request)
  (when (string= request "/")
    (setq request "/top"))  
  (call-handler *lispweb-type-table* request))

(defmethod http-reply ((method (eql :head)) request)
  ;(http-log "HEAD request received")
  (http-reply :get request))

;;; This method should honour the following requests:
;;; /lisp - compile and load a lisp file
;;; /file/path - store a generic file under the specified path
;;; /password - change the opeartor password

(defmethod http-reply ((method (eql :put)) request)
  (call-handler *lispweb-put-table* request))

(defmethod http-reply ((method (eql :delete)) request)
  (let ((password (or (header-value :password) "")))
    (if (check-password password)
	(progn)
      (http-header *http-text-header* :code-name :no-auth))))

;;; HTML objects

(defun register-page (url function class)
  (let ((list (cond ((eq class :user)
		     '*user-pages*)
		    ((eq class :system)
		     '*system-pages*)
		    ((symbolp class)
		     class)
		    (t (warn "~s is not a class name." class)))))
    (set list (add-obj (symbol-value list) url function))))

;;; Image maps

(defun point-in-area-p (x y region)
  "Test whether the point (x, y) is inside area. Area can be:
:rect x1 y1 x2 y2
:circle xc yc r
:default"  
  (case (car region)
    (:default t)
    (:rect (apply #'(lambda (x1 y1 x2 y2)
		      (and (< x1 x x2)
			   (< y1 y y2)))
		  (cdr region)))
    (:circle (apply #'(lambda (xc yc r)
			(let ((dx (- x xc))
			      (dy (- y yc)))
			  (< (+ (* dx dx)
				(* dy dy))
			     (* r r))))
		    (cdr region)))))

;;; URL mapping

(defun url (symbol &key (full t) args)
  "Returns the full url of a page."

  (declare (special lispweb-page-key))
  (when (keyed-class-p (html::page-class symbol))
    (push lispweb-page-key args))
  
  (let ((path (get symbol :lispweb-path)))
    (if path 
	(let ((url (if full
		       (let ((class (get symbol :lispweb-class)))
			 (if class 
			     (concatenate 'string (url class) path)
			   path))
		     (subseq path 1))))
	  (lw-log :url "Link to page ~a~%" symbol)
	  (if args 
	      (format nil "~a~{/~a~}" url args)
	    url))
      (unless (keywordp symbol)
	(lw-log :url "Non-existent page ~a~%" symbol)))))

(defun set-symbol-info (symbol type path class)
  (setf (get symbol :lispweb-fun-type) type)
  (setf (get symbol :lispweb-path) path)
  (setf (get symbol :lispweb-class) class))

;;; URL parsing

(defun url-to-list (url)
  "Turn a URL into a list: (host port path).
Works only for HTTP urls."
  (let ((host nil)
	(port 80)
	(path "/")
	(first-slash (position #\/ url))
	(first-colon (position #\: url)))
    (when (and (> (length url) 7)
	       (string= (subseq url 0 7) "http://"))
      (setq url (subseq url 7))
      (setq first-colon (position #\: url))
      (setq first-slash (position #\/ url)))
    (if first-colon
	(progn 
	  (setq host (subseq url 0 first-colon))
	  (setq port (parse-integer (subseq url (1+ first-colon) first-slash))))
      (setq host (subseq url 0 first-slash)))
    (when first-slash
      (setq path (subseq url first-slash)))
    (list host port path)))

(defun list-to-url (list)
  "Turn a list (host port path) into an HTTP URL."
  (let ((host (car list))
	(port (second list))
	(path (third list)))
    (concatenate 'string 
      "http://" 
      host
      (if (= port 80)
	  ""
	(format nil ":~d" port))
      (if (eq (aref path 0) #\/)
	  path
	(concatenate 'string "/" path)))))
