;;; -*- Mode:Lisp; Base:10; Package: http -*-

;;; File: http.lisp
;;; Program: 

;;; Author: Alberto Riva (alb@ipvaim.unipv.it)
;;; Created: Tue May  9, 1995
;;; Updated: Wed Nov 22, 1995

;;; Contents:
;;;   Variable    *STSP-TASKS* 
;;;   Structure   TASK 
;;;   Function    PRINT-TASK 
;;;   Structure   TASK-FUNCTION 
;;;   Function    PRINT-TASK-FUNCTION 
;;;   Function    DEFINE-TASK 
;;;   Function    FIND-TASK 
;;;   Function    FIND-TASK-FUNCTION 
;;;   Macro       DEFINE-TASK-FUNCTION 
;;;   Function    EXECUTE-TASK-FUNCTION 
;;;   Structure   STSP-AGENT 
;;;   Function    PRINT-AGENT 
;;;   Function    STSP-REPLY 
;;;   Function    STSP-HEADER 
;;;   Function    STSP-ERROR 
;;;   Function    STSP-REDIRECT 
;;;   Page        "TASK 
;;; -*- -*-

(in-package :stsp)

;;; This file implements a simple Task Description Language to be used
;;; in STSP requests.

(defconstant *stsp-version* "0.1")

(defvar *stsp-tasks* nil
  "The list of stsp tasks known to the server.")
  
(defstruct (task (:print-function print-task))
  name
  description
  actions				;a list of symbols
  interface-file			;lisp files that contain the task's public interface
  source-files				;list of lisp files that contain the task implementation
  authinfo)				;who can use this task

#|
authinfo will contain a description of the operations permitted to
other agents. main classes are:  use the task (become a remote user), download 
the public interface (become a local user) and download the implementation (become
an equivalent provider agent).
|#

(defun print-task (task stream ignore)
  (declare (ignore ignore))
  (format stream "<Task ~s>" (task-name task)))

(defstruct (task-action (:print-function print-task-action))
  name
  arglist				;for documentation purposes
  description
  task					;name of the parent task
  code					;the function
  authinfo)				;who can use this function

(defun print-task-action (action stream ignore)
  (declare (ignore ignore))
  (format stream "<Action ~s of task ~s>" (task-action-name action) (task-name (task-action-task action))))

(defun define-task (name &optional (description ""))
  "Define a new task."
  (let ((the-task (find-task (string name))))
    (if the-task
	(setf (task-name the-task) name
	      (task-description the-task) description)
	(push (make-task :name name
			 :description description)
	      *stsp-tasks*))))

(defun find-task (name)
  "Retrieve a task given its name."
  (find-if #'(lambda (task)
	       (equalp (string (task-name task))
		       (string name)))
	   *stsp-tasks*))

(defun find-task-action (task name)
  "Retrieve a action from a task given its name."
  (find-if #'(lambda (func)
	       (equalp (string (task-action-name func))
		       (string name)))
	   (task-actions task)))

(defmacro define-task-action ((task-name action-name &optional (description ""))
			      arglist
			      &body body)
  "This macro creates a action associated with a task. It returns NIL if the task doesn't exist."
  `(let ((the-task (find-task (string ,task-name)))
	 (code-name (intern (format nil "[~a]~a" ,task-name ,action-name) :http)))
    (if the-task
	(let ((the-action (find-task-action the-task (string ,action-name))))
	  (when the-action
	    (setf (task-actions the-task)
		  (remove the-action (task-actions the-task))))
	  (compile code-name '(lambda ,arglist ,@body))
	  (push (make-task-action :name ,action-name
				  :description ,description
				  :arglist ',arglist
				  :task the-task
				  :code (symbol-function code-name))
		(task-actions the-task))))))

#|
(defun execute-task-action (task action arguments)
  (let ((the-action (find-task-action task action)))
    (if the-action
	(handler-case
	    (apply (task-action-code the-action) arguments) ;the action will put its own header
	  (simple-error (c)
					;#+lucid   (apply #'format t (simple-condition-format-string c) (simple-condition-format-arguments c))
					;#+allegro (apply #'format t (simple-condition-format-control c) (simple-condition-format-arguments c))
	    (stsp-error :error "internal-server-error")))
      (stsp-error :not-found action))))
|#

(defun execute-task-action (task action arguments)
  (let ((the-action (find-task-action task action)))
    (if the-action
	(apply (task-action-code the-action) arguments) ;the action will put its own header
      (stsp-error :not-found action))))

(defstruct (stsp-agent (:print-function print-agent))
  name
  location
  port
  crypt-key
  task-info
  credit)

(defun print-agent (agent stream ignore)
  (declare (ignore ignore))
  (format stream "<Agent ~a>" (stsp-agent-name agent)))

;;; The STSP protocol (server side)

(defmethod http-reply ((method (eql :stsp)) request)
  "Parse an STSP request and call the appropriate task action."

  (let ((req-list (explode-request request)))
    (if (< (length req-list) 2)
	(stsp-error :bad-syntax request)
      (let ((task-name (read-from-string (car req-list) nil nil))
	    (action-name (read-from-string (second req-list) nil nil))
	    (action-arguments (cddr req-list)))
	(let ((task (find-task task-name)))
	  (http-debug-message "~s ~s ~s" task-name action-name action-arguments)
	  (if task
	      (execute-task-action task action-name action-arguments)
	    (stsp-error :not-found task-name)))))))
	      
#|  
  (multiple-value-bind (task-name task-end)
      (read-from-string request nil nil)
    (if task-name
	(multiple-value-bind (action-name action-end)
	    (read-from-string request nil nil :start task-end)
	  (if action-name
	      (let ((arguments (read-from-string request nil nil :start action-end))
		    (task (find-task task-name)))
		(http-debug-message "~s ~s ~s" task-name action-name arguments)
		(if task
		    (execute-task-action task action-name arguments)
		    (stsp-error :not-found task-name)))
	      (stsp-error :bad-syntax request)))
	(stsp-error :bad-syntax request))))
|#
	
(defun stsp-header (code-name &optional additional-headers)
  (let ((code (if (numberp code-name)
		  code-name
		(assoc code-name http::*status-code-alist*))))
    (setf (http::http-info-code *http-info*) code-name)
    (format nil "HTTP/1.0 ~d ~a~%~
~aDate: ~a
~:{~:(~a~): ~a~%~}~%"
	    (second code) 
	    (third code) 
	    *LispWeb-base-header*
	    (http-time) 
	    additional-headers 
	    #\Return)))

;;; A couple of specialized headers

(defun stsp-error (code-name object)
  (format *http-stream* (stsp-header code-name (list (list :Error-object object)))))

(defun stsp-redirect (location)
  (format *http-stream* (stsp-header :redirect (list (list :Location location)))))

;;; LispWeb documentation

(define-page ("Task description" :func-name task-info :class :system)
    (task-name &optional action-name)
  (let* ((the-task (find-task task-name)))
    (html-out *lispweb-banner*)
    (if action-name
	(let ((the-action (find-task-action the-task action-name)))
	  (html-out (center (header 1 (html-build "The ~:(~a~) action~%<P>" action-name))))
	  (html-out "<P>The argument list for this action is: ~a~%<P>" (bold (format nil "~:(~a~)" (task-action-arglist the-action))))
	  (html-out (italic (task-action-description the-action)))
	  (html-out "<P>This action belongs to the ~a task.~%" (anchor (format nil "/task-info/~a" task-name) task-name)))	  
	(progn
	  (html-out (center (header 1 (html-build "The ~:(~a~) task~%<P>" task-name))))
	  (html-out (italic (task-description the-task)))
	  (html-out "<P>This task provides the following actions:~%")
	  (html-out (dotted-list (mapcar #'(lambda (func)
					     (format nil "~a  ~(~a~)"
						     (anchor (format nil "/task-info/~a/~a" task-name (task-action-name func))
							     (format nil "~:(~a~)" (task-action-name func)))
						     (task-action-arglist func)))
					 (task-actions the-task))))
	  (html-out "<HR>Back to the ~a." (anchor "/" "top page"))))))

;;; The STSP protocol (client side)

(defun format-headers (headers)
  (format nil "~{~@(~a~): ~a~%~}" headers))

#|
(defmacro with-stsp (host port request headers &body body)
  `(let ((*http-stream* (tcp::telnet ,host ,port)))
     (unwind-protect
	 (let ((*stsp-headers* "")
	       (line ""))
	   (format *http-stream* "STSP ~a HTTP/1.0~%~%" ,request)
	   (format *http-stream* "~a~%" (format-headers ,headers))
	   (force-output *http-stream*)
	   (setq *connection-header* (http-read-line *http-stream*))
	   (let ((http-code (read-from-string *connection-header* nil nil :start 9)))
	     (loop 
	      (when (string= (setq line (http-read-line *http-stream* "")) "")
		(return)))
	     (setq *connection-header* (concatenate 'string *connection-header* (string #\Newline) line))
	     ,@body))
      (tcp::close-tcp-stream *http-stream*))))
|#

(defmacro with-stsp (host port request headers &body body)
  (let ((stream-name (gensym)))
    `(let ((,stream-name (tcp::telnet ,host ,port))
	   (*http-info* (make-http-info)))
       (unwind-protect
	   (progn
	     (setf (http-info-stream *http-info*) ,stream-name)
	     (format ,stream-name "STSP ~a HTTP/1.0~%" ,request)
	     (format ,stream-name "~a~%" (format-headers ,headers))
	     (http::http-force-output ,stream-name)
	     (http::process-response-line)
	     
	     (let* ((code (http::http-info-code *http-info*))
		    (status (http::code-to-status code)))
	       (when (>= code 300)
		 (if status 
		     (error "STSP Error: ~a" (third status))
		   (error "STSP Unidentified error"))))
	     
	     ;; call error handlers if code <> 200
	     
	     (setf (http-info-headers *http-info*)
	       (process-headers ,stream-name))
	     ,@body)
	 (close ,stream-name)))))
  
;;; Definition of the :system task. Every STSP server must implement this.

(define-task :system
    "The basic STSP task. It implements the basic functionalities needed by an STSP agent.")

(define-task-action (:system :version "Return the version of this STPS agent.")
    ()
  (format *http-stream* (stsp-header :ok))
  (format *http-stream* "~a~%" *stsp-version*))

(define-task-action (:system :codes "Return the return codes produced by this STSP agent.")
    ()
  (format *http-stream* (stsp-header :ok))
  (mapc #'(lambda (code-line)
	    (format *http-stream* "~d ~a~%" (second code-line) (third code-line)))
	*status-code-alist*))

(define-task-action (:system :tasks "Return the tasks known to this STSP agent.")
    ()
  (format *http-stream* (stsp-header :ok))
  (format *http-stream* "~{~a ~}~%" (mapcar #'task-name *stsp-tasks*)))

(define-task-action (:system :actions "Return the actions in task task-name.")
    (task-name)
  (let ((the-task (find-task task-name)))
    (cond (the-task
	   (format *http-stream* (stsp-header :ok))
	   (format *http-stream* "~{~a ~}~%" (mapcar #'task-action-name (task-actions the-task))))
	  (t (stsp-error :not-found task-name)))))

(define-task-action (:system :arglist "Return the argument list of action action-name of task task-name.")
    (task-name action-name)
  (let ((the-task (find-task task-name)))
    (if the-task
	(let ((the-action (find-task-action the-task action-name)))
	  (cond (the-action
		 (format *http-stream* (stsp-header :ok))
		 (format *http-stream* "~s~%" (task-action-arglist the-action)))
		(t (stsp-error :not-found action-name))))
	(stsp-error :not-found task-name))))

(define-task-action (:system :identity "Return the identity of this STSP agent.")
    ()
  (format *http-stream* (stsp-header :ok))
  (format *http-stream* "~a~%~a~%" (machine-instance) "My PGP Key."));these should come from the agent's fields

;;; Lisp client interface for the :system task

(defun stsp-version (host port)
  (with-stsp host port "system version"
    (http-read-line *http-stream* :unknown)))

(defun stsp-codes (host port &aux line)
  (with-stsp host port "system codes"
    (loop
     (unless (setq line (http-read-line *http-stream* nil))
       (return))
     (print line))))

(defun stsp-tasks (host port)
  (with-stsp host port "system tasks"
    (http-read-line *http-stream* nil)))
