;;; -*- Mode: LISP; Syntax: Common-lisp; Base: 10; Package: OCML;   -*-

(in-package "OCML")

(in-ontology gen-design)

          
;;;DESIGN-OPERATOR
(def-class DESIGN-OPERATOR ()
  "A state transition in the state space model specifies a link between 
   two design states (in practice two design models).  State transitions are 
   carried out by means of design operators."
  ((assumption :default-value (true)
               :type relation-expression
               :documentation 
               "This slot can be used to specify a statement that is expected
                to hold for the application domain where the operator is applied.  
                The difference between assumptions and preconditions is that the former
                are static and the latter are dynamic.  The truth value of a precondition
                might change during the design process.  Assumptions are expected to remain
                (un-)satisfied during the design process")
   
   (has-body :type design-operator-body))
  :Lisp-slots ((effective-parameter-list :initform nil
                                         :accessor effective-parameter-list)
               (applicability-env :initform nil :accessor applicability-env))
  :lisp-class-name design-operator)


(def-class DESIGN-OPERATOR-BODY (unary-function) ?fun
  "A design operator body is a unary function which takes as input a design model,
   say Di, and produces as output a design model Dj such that Di = Dj"
  :no-op (:constraint (and (domain ?fun design-model)
                           (range ?fun design-model)
                           (<=> (= (call ?fun ?di) ?dj)
                                (not (= ?di ?dj))))))

(def-class basic-operator (design-operator)
  "A class of operators which can be applied only once")

(def-class multiple-operator (design-operator)
  "A class of operators which can be applied more than once")

  
(def-class DESIGN-EXTENSION-OPERATOR (design-operator)
  "A class of operators which are used to provide values
   for unbound parameters")

(def-class DESIGN-EXTENSION-OPERATOR (design-operator) 
  "A basic design extension operator is the simplest kind 
   of design operator.  
   Its body is a unary function which takes as argument 
   an unbound  parameter, ?p, and a design model, and produces
    as a result a  new value, ?z, which is taken to specify the value of 
   ?p in the next design state.  
   The values of all the other design parameters should 
   not be affected by the application of the operator.  
   This operator is 'basic' in the sense that
   it can only spawn one successor state"
  ((applicable-to-parameters 
    :default-value '(setofall ?x (parameter ?x))
    :type function-expression
    :documentation "An expression which returns the set
                    of parameters whose value can be computed 
                    by means of this operator")
   (has-precondition 
    :default-value (kappa (?d) (true))
    :type relation-expression
    :documentation "This is an expression which can be used to determine 
                    whether an operator is applicable in the current 
                    state - i.e. design model.  This expression
                    should not depend on a particular design 
                    parameter")
 (has-body :type design-extension-operator-body))
  :lisp-class-name design-extension-operator)
  

(def-class DESIGN-EXTENSION-OPERATOR-BODY (lambda-expression) ?x
  "A basic design extension operator body is a unary function which takes an  
   unbound parameter, say ?p and produces a result, ?z, which belongs to the 
   value range of ?.  ?z is taken as the new value of ?p in the successor 
   design state"
  :no-op (:constraint (and (nth-domain ?x 1 parameter)
                           (nth-domain ?x 2 ?dm)
                           (=> (= ?z (call ?x ?p))
                               (and (has-value-range ?p ?range)
                                    (member ?z ?range))))))

(def-class basic-DESIGN-EXTENSION-OPERATOR (design-EXTENSION-operator basic-operator)
  ()
  :lisp-class-name basic-design-extension-operator)


(def-class multiple-DESIGN-EXTENSION-OPERATOR (design-extension-operator multiple-operator) 
  "The body of a  multiple design extension operator is a binary  
   function which takes as argument an unbound  parameter, ?p, and a set of 
   values, ?values, and produces as a result a new value, ?z, 
   which is taken to specify the value of ?p in the next design state.  
   The values of all the other design parameters should not be affected by 
   the application of the operator.  A multiple design operator can 
   be invoked repeatedly to spawn alternative successor states"
  (
   (has-precondition :default-value (kappa (?d) (true))
                     :type relation-expression
                     :documentation "This is an expression which can be used to determine 
                                     whether an operator is applicable in the current 
                                     state - i.e. design model.  This expression
                                     should not depend on a particular design 
                                     parameter")
   (has-body :type multiple-design-extension-operator-body))
  :lisp-class-name multiple-design-extension-operator)

(def-class multiple-design-extension-operator-body (lambda-expression) ?x
  "A multiple design extension operator body is a binary function which takes a 
   parameter, say ?p, and a list of values, say ?values, and produces a result, 
   ?z, which belongs to the value range of ?p but is not a member of the 
   list ?values.  ?z is taken as the new value of ?p in the successor design state"
  :no-op (:constraint (and (nth-domain ?x 1 parameter)
                           (nth-domain ?x 2 ?y)
                           (=> (= ?z (call ?x ?p ?values))
                               (and (has-value-range ?p ?range)
                                    (forall ?v (=> (member ?v ?values)
                                           (member ?v ?range)))
                                    (member ?z ?range)
                                    (not (member ?z ?values)))))))





;;;DESIGN-OPERATOR-ORDER 
(def-relation DESIGN-OPERATOR-ORDER (?x ?C)
  "This relation can be used to specify an application specific ordering
   of operators")

(def-rule design-operator-transitivity
  ((design-operator-order ?x ?C)
   if
   (asserted (design-operator-order ?x ?c2))
   (<> ?c ?c2)
   (design-operator-order ?c2 ?c)))


(def-relation PRECONDITION-HOLDS (?op ?dm)
  :constraint (and (design-operator ?op)
                   (design-model ?dm))
  :iff-def (and (has-precondition ?op ?exp)
                (holds ?exp ?dm)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmethod  OPERATOR-STILL-USEFUL? ((op design-operator) d)
  "An operator is still useful if some of the parameters which can 
   be assigned a value by it are still unbound"
  (some #'(lambda (p)
            (unbound? p d))
        (effective-parameter-list op)))

;;;COMPUTE-EFFECTIVE-APPLICABILITY-SET
(defmethod  compute-effective-applicability-set ((op design-operator))
  (ocml-eval-fun (the-slot-value (name op) 'applicable-to-parameters)))

;(defmethod  PRECONDITION-HOLDS? ((op design-operator) dm)
;  (holds? `(holds ,(the-slot-value (name op) 'has-precondition)
;                  ',dm)))

(defmethod  PRECONDITION-HOLDS? ((op design-operator) dm)
  (goal-holds? `(precondition-holds ,(name op) ',dm)))

