;;; -*- Mode: LISP; Syntax: Common-lisp; Base: 10; Package: OCML;   -*-

;;;This file contains the definitions linking the generic problem
;;;solving model to specific design knowledge.

(in-package "OCML")

(defvar *design-constraints* nil)
(defvar *design-parameters* nil)
(defvar *design-requirements* nil)
(defvar *design-operators* nil)
(defvar *cost-function* nil)
(defvar *cost-order-rel* nil)
(defvar *assignment-rel* nil)
(defvar *trace-design-process?* t)
(defvar *current-state* nil)



(defun cache-design-specification (appl method)
  (let ((task (the-slot-value appl 'tackles-task)))
    (setf *design-operators* 
        (mapcar #'(lambda (x)
                    (find-instance x 'design-operator))
                (the-slot-value (name method) 'has-design-operators))
        *design-parameters* 
        (mapcar #'(lambda (x)
                    (find-instance x 'parameter))
                (the-slot-value task 'has-parameters))
        *design-constraints* 
        (mapcar #'(lambda (x)
                    (find-instance x 'constraint))
                (the-slot-value task 'has-constraints))
        *design-requirements* 
        (mapcar #'(lambda (x)
                    (find-instance x 'requirement))
                (the-slot-value task 'has-requirements))
        *assignment-rel* (the-slot-value 
                          (the-slot-value task 'has-design-model)
                          'membership-test)
        *cost-function* (the-slot-value task
                                        'has-cost-function)
        *cost-order-rel* (the-slot-value task
                                        'has-cost-order-relation))))

;;;CACHE-PARAM-TO-OPS-LINKS
(defun cache-param-to-ops-links ()
  (loop for op in *design-operators*
        for params = (compute-effective-applicability-set op)
        do
        (loop for param in params
              do
              (link-param-to-operator (find-instance param 'parameter)
                                      op))))

;;;LINK-PARAM-TO-OPERATOR
(defmethod link-param-to-operator ((param parameter) (op design-operator))
    (with-slots (effective-parameter-list) op
      (pushnew param effective-parameter-list)
      (pushnew op  (relevant-operators param))))



;;;*DESIGN-SPACE*
(defvar *design-space* nil)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;APPLY-METHOD - An application is executed by applying the chosen method 
;;;to the instance representing the application.  This definition ensures that for
;;;design methods this call results in a call to top-level-design.
(defmethod APPLY-METHOD ((method design-method) appl)
  (top-level-design appl method ))

(defmethod TOP-LEVEL-DESIGN (appl (method design-method))
  (cache-design-specification appl method)
  (cache-param-to-ops-links)
  (initialise-design-space appl method)
  (catch 'exit
    (loop
      (let ((state (select-design-state appl method *design-space*)))
        (make-this-state-current appl state method)
        (design-from-state appl method state)))))


;;;METHOD INITIALISE-DESIGN-SPACE
;;;Generates the starting point in the search space
(defmethod initialise-design-space (appl (method design-method))
  (setf *design-space* nil)
  (add-design-state appl method nil nil))

(defun INITIAL-STATE ()
  (car *design-space*))


(defmethod ADD-DESIGN-STATE (appl (method design-method) predecessor extension)
  (let ((state (new-design-state appl method predecessor extension)))
    (evaluate-state appl method state)
    (internal-add-state method state)))

(defmethod REMOVE-DESIGN-STATE ((state design-state))
  (setf *design-space*
        (remove state *design-space*)))

(defmethod NEW-DESIGN-STATE (appl (method design-method)
                                  &optional  predecessor extension)
  appl
  (cond (predecessor
         (make-instance (design-state-type method)
           :design-model (cons extension (design-model predecessor))
           :design-extension extension
           :depth (1+ (state-depth predecessor))
           :predecessor predecessor
          ))
        (t
         (make-instance (design-state-type method)
           :design-model nil))))


(defmethod internal-add-state ((method design-method) (state design-state))
  (setf *design-space* (cons state *design-space*)))


(defmethod MAKE-THIS-STATE-CURRENT (appl (state design-state)(method design-method))
  appl
  (unless (eq state *current-state*)
    (with-slots (design-model) state
      (loop for pair in (setofall '(?x . ?C)`(,*assignment-rel* ?x ?C ))
            unless (member pair design-model :test #'equal)
            do
            (unassert1 `(,*assignment-rel* ,(car pair) ,(cdr pair))))
      (loop for assignment in design-model
            do
            (tell1 `(,*assignment-rel* ,(car assignment) ,(cdr assignment)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defmethod SELECT-DESIGN-STATE (appl (method design-method) space)
  appl method

  (when *trace-design-process?*
    (when space
      (lines 2)
      (print-with-spaces (state-depth (car space)) 
                         "Current state is now ~s"
                         (car space))))

  (car space))


(defmethod SELECT-FOCUS (appl (method design-method) (state design-state) 
                              (focus-type (eql :parameter)))
  (let ((focus (car  (state-parameters state))))
    appl method
   (when *trace-design-process?*
     (if focus
       
             (print-with-spaces (state-depth state)
                                "Focus on ~a" (name (car  (state-parameters state))))
             (print-with-spaces (state-depth state)
                                "Cannot select any focus")))
   focus))


(defmethod EVALUATE-STATE (appl (method design-method)(state design-state))
  (if (eq *current-state* state)
    (evaluate-state-internal appl method state)
      (Let ((s *current-state*))
        (make-this-state-current appl state method)
        (evaluate-state-internal appl method state)
        (when s
          (make-this-state-current appl s method)))))

(defmethod EVALUATE-STATE-INTERNAL (appl (method design-method)(state design-state))
   (setf (state-feasible? state) (evaluate-feasibility appl method state)
        (state-parameters state) (collect-state-parameters appl state)
        (state-operators state)(collect-applicable-operators appl method state) 
        (violated-constraints state) (evaluate-constraints appl method
                                                           state)
        
        (state-complete? state) (design-complete? appl method state)
        (state-cost state) (evaluate-state-cost appl method state)))


      
(defmethod EVALUATE-FEASIBILITY (appl (method design-method)(state design-state))
  (declare (ignore appl method state))
  "This is just to provide a default"
  t)

(defmethod EVALUATE-CONSTRAINTS (appl (method design-method)(state design-state))
  appl
  (loop with design-model = (design-model state)
        with constraints-violated
        for c in *design-constraints*
        for flag = (constraint-violated? c design-model)
        when flag
        do
        (push c constraints-violated)
        finally

        (when (and *trace-design-process?*
                   constraints-violated)
          (print-with-spaces (state-depth state)
                             "The following constraints have been violated: ~{~A ~}"
                  constraints-violated))

        (return constraints-violated)))

(defmethod filter-applicable-operators ((state design-state) ops)
  (loop with applicable-operators
        for op in ops
        for result = (precondition-holds? op)
        when result
        do
        (push op applicable-operators)
        (setf (applicability-env op) (when (listp result) result))
        finally
        (return  applicable-operators)))


(defmethod collect-applicable-operators (appl (method design-method)
                                                   (state design-state))
  appl
  (filter-applicable-operators 
   state
   (apply #'append 
          (mapcar #'relevant-operators 
                  (state-parameters state)))))
    
(defmethod COLLECT-STATE-PARAMETERS (appl (state design-state))
  appl
  (if (predecessor state)
    (remove (find-instance (car (design-extension state))'parameter)
            (state-parameters (predecessor state)))
    *design-parameters*))

(defmethod DESIGN-COMPLETE? (appl (method design-method)
                                  (state design-state))
  appl
  (holds? `(design-model-complete ,(design-model state)
                                  ,(mapcar #'name
                                        *design-parameters*))))
    
        
(defun DESIGN-SUCCEEDS (state)
  state
  
  (when *trace-design-process?*
    (print-with-spaces (state-depth state) 
                       "Design is now complete.  Cost of final state is ~s"
                       (state-cost state)))

  (throw 'exit t))

(defun DESIGN-FAILS (state)
  state
  
  (when *trace-design-process?*
    (print-with-spaces (if state (state-depth state) 0)
                       "Cannot select a state...design fails"))

  (throw 'exit nil))

(defun state-fails (state)
  (when *trace-design-process?*
    (print-with-spaces (state-depth state) 
                       "State ~s fails" state))
  (remove-design-state state))


(defmethod EXTEND-DESIGN (appl (method design-method)  
                               (state design-state)
                               focus)
  
  
  (setf (current-focus state) focus
        (state-parameters state) (remove focus 
                                         (state-parameters state))
        (focus-operators state) (sort-operators appl method state
                                                (filter-operators-applicable-to-focus 
                                                 method state focus)))
  (cond ((focus-operators state)
         (design-from-focus appl method state focus))
        (t
         (backtrack-from-focus appl method state focus))))

(defmethod DESIGN-FROM-FOCUS (appl (method design-method)  
                                   (state design-state)
                                   focus)
  (let ((design-op (select-design-operator appl method state focus 
                                           (focus-operators state))))
    (cond (design-op
           (setf (current-operator state) design-op
                 (nogood-values state) nil
                 (focus-operators state) 
                 (remove 
                  design-op 
                  (focus-operators state)))
           (try-design-operator appl method design-op state focus))
          (t
           (backtrack-from-focus appl method state focus)))))


(defmethod BACKTRACK-FROM-FOCUS (appl 
                                 (method design-method) 
                                 (state design-state) 
                                 (focus parameter))
  (when *trace-design-process?*
    (print-with-spaces (state-depth state)
                       "No more operators applicable to ~a...trying another focus" 
            (name focus)))
  (if (state-parameters state)
    (let ((focus (select-focus appl method state :parameter)))
          (if focus
            (extend-design appl method state focus)
            (state-fails state)))
    (state-fails state)))

(defmethod FILTER-OPERATORS-APPLICABLE-TO-FOCUS ((method design-method) 
                                                (state design-state)  
                                                (focus parameter))
  (intersection (state-operators state)
                (relevant-operators focus)))


(defmethod SORT-OPERATORS (appl (method design-method) 
                                (state design-state)
                                ops)
  appl
  (mapcar #'find-instance
          (ocml-eval-fun `(sort ',(mapcar #'name ops)
                                design-operator-order))))

(defmethod SELECT-DESIGN-OPERATOR (appl (method design-method) 
                                   (state design-state)
                                   focus
                                   ops)
  (declare (ignore   focus method appl))
  (let ((op (car ops)))
    (when *trace-design-process?*
      (if op
        (print-with-spaces (state-depth state)
                           "Current operator is ~a" (name op))
        (print-with-spaces (state-depth state)
                           "Cannot select an operator")))
    op))

(defmethod RETRY-DESIGN-OPERATOR (appl (method design-method)
                                     (design-op design-operator)
                                     (state design-state)  
                                     (focus parameter))
    (when *trace-design-process?*
      (print-with-spaces (state-depth state)
                         "Retrying ~a on ~a" 
                         (name design-op)
                         (name focus)))
  (try-design-operator 
   appl method design-op state focus))


(defmethod TRY-DESIGN-OPERATOR (appl (method design-method)
                                     (design-op design-operator)
                                     (state design-state)  
                                     (focus parameter))
  (multiple-value-bind (value cont) 
                       (apply-design-operator 
                        design-op state  focus)
    (setf (continuation state) cont)
    (cond ((eq value :nothing)
           (backtrack-from-operator appl method state design-op focus))
          (t
           (setf (nogood-values state)
                 (cons value (nogood-values state)))
           (when *trace-design-process?*
               (print-with-spaces (state-depth state)
                                  "New assignment is <~a ~A>" (name focus) value))
           (add-design-state appl  method state  (cons (name focus) value))))))

(defmethod BACKTRACK-FROM-OPERATOR (appl 
                                    (method design-method) 
                                    (state design-state) 
                                    (design-op design-operator)
                                    (focus parameter))
  (when *trace-design-process?*
    (print-with-spaces (state-depth state)
                       "Operator ~a fails..trying another one..." (name design-op)))
  (design-from-focus appl method state focus))


(defmethod  APPLY-DESIGN-OPERATOR ((design-op design-operator)
                                  (state design-state)  
                                  (focus parameter))
  (with-slots (applicability-env) design-op
    (ocml-apply (the-slot-value (name design-op) 'body)
                       (List (name focus)
                             (list 'quote (nogood-values state)))
                       applicability-env)))


;;;;;;;;;;;;;;;;;;;;;

(defmethod EVALUATE-STATE-COST (appl (method design-method)(state design-state))
  appl
  (with-slots (design-model) state
    (ocml-call *cost-function* `',design-model)))


(defmethod cheaper-state? ((state1 design-state)(state2 design-state))
  (let ((c1 (state-cost state1))
        (c2 (state-cost state2)))
    (cheaper-cost? c1 c2)))

(defun cheaper-cost? (c1 c2)
  (holds? `(,*cost-order-rel* ',c1 ',c2)))



          
;;;;;;;;;;;;;;;;;;;;;;;


(def-class PROPOSE&BACKTRACK (design-method)
   :lisp-class-name propose&backtrack)

(defmethod DESIGN-FROM-STATE (appl (method propose&backtrack) (state design-state))
  (if (state-feasible? state)
    (if (violated-constraints state)
      (state-fails state)
      (if (state-complete? state)
        (design-succeeds state)
        (if (state-operators state)
          (if (current-focus state)
            (retry-design-operator 
             appl method (current-operator state)state (current-focus state))
            (let ((focus (select-focus appl method state :parameter)))
              (if focus
                (extend-design appl method state focus)
                (state-fails state))))
          (state-fails state))))
    (state-fails state)))



