;;; -*- Mode: LISP; Syntax: Common-lisp; Base: 10; Package: OCML;   -*-

(in-package "OCML")

(in-ontology gen-design)


(def-class DESIGN-METHOD (problem-solving-method)
  ((applicable-to-task-type :value parametric-design)
   (has-input-role :value has-design-operators)
   (has-design-operators :type design-operator))
  :lisp-slots ((design-state-type :initform 'design-state
                                  :accessor design-state-type
                                  :initarg :design-state-type))
  :lisp-class-name design-method)


(def-class DESIGN-APPLICATION (application)
  :Lisp-class-name design-application)

(def-relation DEPENDS-ON (?p1 ?p2)
  "This relation models parameter dependencies.
   A parameter, p1, depends on  p2 if p2 
   has to be bound in order to compute p1." 
  :constraint (and (parameter ?p1)
                   (parameter ?p2)))

(def-relation AFFECTS (?p1 ?p2)
  "The inverse of depends-on"
  :constraint (and (parameter ?p1)
                   (parameter ?p2))
  :iff-def (depends-on ?p2 ?p1))


(def-class PARAMETER () ?p
  "This modifies the definition given in the parametric  
   design task ontology. It adds the notion of dependency 
   between parameters, plus some symbol-level slots to cache 
   the links between parameters and the relevant operators and 
   constraints"
  ((has-value-range :type set)
   (depends-on :type parameter 
               :documentation 
               "This expresses the dependencies between 
                parameters. A parameter, p1, depends on 
                p2 if p2 has to be assigned in order to 
                compute p1.  We assume there is only one
                way to compute p1")
   (affects :type parameter))
  :iff-def (exists ?task (and (parametric-design ?task)
                              (has-parameters ?task ?l)
                              (member ?p ?l)))
  :lisp-slots ((relevant-operators :accessor relevant-operators
                                   :initform nil)
               (relevant-constraints :accessor relevant-constraints
                                   :initform nil))
  :lisp-class-name parameter)

  
(def-function RELEVANT-OPERATORS (?param)
  :constraint (parameter ?param)
  :body (setofall ?op (and (design-operator ?op)
                           (member ?param 
                              (the ?l (applicable-to-parameters ?op ?l))))))

(def-relation APPLICABLE-TO-PARAMETERS (?x ?l)
  "A relation which associates an object such as a constraint
   or a design operator to a set of parameters to which the object is 
   'applicable'"
  :constraint (and (set ?l)
                   (every ?l parameter)))



(def-relation COMPUTABLE (?param ?dm)
  :iff-def (and (parameter ?param)
                (design-model ?dm)
                (= ?l (setofall ?x (depends-on ?param ?X)))
                (every ?l (kappa (?x)
                                 (bound-parameter ?x ?dm)))))

(def-function ALL-COMPUTABLE-PARAMETERS (?params ?dm)
  :body (setofall ?x (and (member ?x ?params)
                          (unbound-parameter ?x ?dm)
                          (computable ?x ?dm))))
                          
  

(def-class PARAMETRIC-CONSTRAINT (constraint) 
  "A parametric constraint provides a more 'problem solving oriented'
   definition of constraints.  Both the precondition and the expression
   are binary kappa expressions, which take as argument a parameter and a 
   value (i.e. an assignment).  The idea here is that this assignment-oriented
   approach makes it possible to limit the number of constraints to be tested
   at each stage of the design process to those relevant to the new 
   assignments"
  ((applicable-to-parameters 
    :type function-expression
    :documentation "An expression which returns the set
                    of parameters to which this constraint 
                    is applicable")
   (has-expression 
    :cardinality 1 
    :type legal-parametric-constraint-expression)
   (has-precondition 
    :default-value '(kappa (?p ?d)
                     (true))
    :type kappa-expression
    :documentation "This is an expression which can be used to determine whether
                    a constraint makes sense for a given parameter
                    assignment"))
  :lisp-class-name parametric-constraint
  :Lisp-slots ((effective-parameter-list 
                :initform nil
                :accessor effective-parameter-list)))


(def-function RELEVANT-CONSTRAINTS (?param)
  :constraint (parameter ?param)
  :body (setofall ?c (and (parametric-constraint ?c)
                          (member ?param 
                              (the ?l (applicable-to-parameters ?c ?l))))))

(def-class LEGAL-PARAMETRIC-CONSTRAINT-EXPRESSION (kappa-expression) 
  ?exp
  "This is an expression parametrized over a parameter and a value"
  :iff-def (and (kappa-expression ?exp)
                (== ?exp (kappa ?schema ?sentence))
                (= ?vars (all-free-vars-in-sentence ?sentence))
                (null (difference ?schema ?vars ))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;UNBOUND? --This takes a parameter object (not a name!) and checks
;;;whether it is bound in the current design model
(defmethod unbound? ((param parameter) dm)
  (not (bound? param dm)))

(defmethod BOUND? ((param parameter)dm)
  (goal-holds? `(has-value ,(name param) ?X ,dm)))


(defmethod computable? ((p parameter) design-model)
  (every #'(lambda (param)
             (assoc param design-model))
         (get-slot-values p 'depends-on)))

(defun All-unbound-computable-parameters (parameters design-model)
  (filter parameters #'(lambda (p) (and (not (assoc (name p) design-model))
                                        (computable? p design-model)))))
                       

(defmethod  compute-effective-applicability-set ((c parametric-constraint))
  (ocml-eval-fun (the-slot-value (name c) 'applicable-to-parameters)))

(defmethod  CONSTRAINT-PRECONDITION-HOLDS? ((c parametric-constraint) param value)
  (goal-holds? `(holds ,(the-slot-value (name c) 'has-precondition) 
                  ',param
                   ',value)))

(defmethod ASSIGNMENT-VIOLATES-CONSTRAINT? ((c parametric-constraint) param value )
  (not (goal-holds? `(holds ,(the-slot-value (name c) 'has-expression) 
                  ',param
                   ',value))))

(defun CHECK-DM-CONSTRAINTS (dm)
  "Checks that the input design model satisfies all relevant 
   constraints"
  (loop with constraints-violated
        with bad-assignments
        for pair in dm
        for param = (car pair)
        for value = (cdr pair)
        for relevant-constraints = (relevant-constraints (find-current-instance param 'parameter))
        do
        (loop for c in relevant-constraints
              for flag = (constraint-precondition-holds? c param value)
              when flag
              do
              (setf flag (assignment-violates-constraint? c param value))
                (when flag
                  (pushnew c constraints-violated)
                  (push (cons  c (cons param value)) bad-assignments)))
          finally
          (if ;;;;;(and *trace-design-process?*
                   constraints-violated ;;;;)
            (print-with-spaces 2 ;;; (state-depth state)
                               "The following constraints have been violated: ~{~A ~}~%"
                               constraints-violated)
            (when *trace-design-process?*
              (print-with-spaces 2 ;;; (state-depth state)
                                 "...Done")))
      (return constraints-violated)))
  
  


