package org.infrawebs.axiom.logexpr;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.infrawebs.axiom.model.AxiomContentsModel;
import org.infrawebs.axiom.model.ModelWithConnections;
import org.infrawebs.axiom.model.OperatorModel;
import org.infrawebs.axiom.operations.ModelOperations;

public interface ComplexOperator
{
	OperatorModel getOperator();
	String getPrimaryVariableName();
	void compile();
}

/**
 * An wrapper for NOT operator. It has a single ComplexConjunction operand. 
 */
class ComplexNegation implements ComplexOperator
{
	OperatorModel operator = null;
	ComplexConjunction operand = null;

	public ComplexNegation(AxiomContentsModel contents) {
		this.operator = contents.getFactory().createOperator(OperatorModel.OPERATOR_NOT, null);
	}

	public OperatorModel getOperator() {
		return this.operator;
	}

	public ComplexConjunction getOperand() {
		return this.operand;
	}

	public void setOperand(ComplexConjunction operand) {
		this.operand = operand;
	}

	/**
	 * Compiles the operand and connects it to the operator.
	 * @see ComplexConjunction.compile().
	 */
	public void compile() {
		ModelWithConnections top = this.operand.compile();
		ModelOperations.connect(this.operator, top);
	}

	/**
	 * @return The operand's primary variable name.
	 * @see ComplexConjunction.getPrimaryVariableName().
	 */
	public String getPrimaryVariableName() {
		return this.operand.getPrimaryVariableName();
	}
}

/**
 * An wrapper for the OR operator. It has a list of ComplexConjunction operands. 
 */
class ComplexDisjunction implements ComplexOperator
{
	OperatorModel operator = null;
	List<ComplexConjunction> operands = new ArrayList<ComplexConjunction>(2);

	public ComplexDisjunction(AxiomContentsModel contents) {
		this.operator = contents.getFactory().createOperator(OperatorModel.OPERATOR_OR, null);
	}

	public OperatorModel getOperator() {
		return this.operator;
	}

	public List listOperands() {
		return this.operands;
	}

	public void addOperand(ComplexConjunction operand) {
		this.operands.add(operand);
	}

	/**
	 * Compiles all operands and connects them to the operator.
	 * @see ComplexConjunction.compile().
	 */
	public void compile() {
		Iterator i = this.operands.iterator();
		while (i.hasNext()) {
			ComplexConjunction conjunction = (ComplexConjunction)i.next();
			ModelWithConnections top = conjunction.compile();
			ModelOperations.connect(this.operator, top);
		}
	}

	/**
	 * @return A variable name if all operands represent the same variable.
	 *         Null if any two operands represent different variables.
	 */
	public String getPrimaryVariableName() {
		Iterator i = this.operands.iterator();
		String result = null;
		while (i.hasNext()) {
			String varName = ((ComplexConjunction)i.next()).getPrimaryVariableName();
			if (varName == null) {
				return null;
			}
			else if (result == null) {
				result = varName;
			}
			else if (!result.equals(varName)) {
				return null;
			}
		}
		return result;
	}
}

/**
 * A dummy class used to separate disjunctions and negations in the operator stack
 * while preprocessing logical expressions.
 */
class DummyConjunction implements ComplexOperator
{
	public static final DummyConjunction instance = new DummyConjunction();
	
	private DummyConjunction() {
		// Just hides the constructor.
	}

	public OperatorModel getOperator() {
		return null;
	}

	public String getPrimaryVariableName() {
		return null;
	}

	public void compile() {
		// Do nothing - this is a dummy implementor.
	}
}

