package org.infrawebs.axiom.logexpr;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.infrawebs.axiom.model.AttributeModel;
import org.infrawebs.axiom.model.AxiomContentsModel;
import org.infrawebs.axiom.model.AxiomRootModel;
import org.infrawebs.axiom.model.InstanceModel;
import org.infrawebs.axiom.model.MemberOfModel;
import org.infrawebs.axiom.model.ModelWithConnections;
import org.infrawebs.axiom.model.OperatorModel;
import org.infrawebs.axiom.model.ParameterModel;
import org.infrawebs.axiom.model.RelationModel;
import org.infrawebs.axiom.model.ValueInstanceModel;
import org.infrawebs.axiom.model.VarNameHolderModel;
import org.infrawebs.axiom.model.VarNameList;
import org.infrawebs.axiom.model.VarNameModel;
import org.infrawebs.axiom.model.VariableModel;
import org.infrawebs.axiom.plugin.AxiomEditorPlugin;
import org.infrawebs.common.Util;
import org.infrawebs.ontology.shells.AttributeShell;
import org.infrawebs.ontology.shells.OntologyShell;
import org.omwg.logicalexpression.Atom;
import org.omwg.logicalexpression.AttributeValueMolecule;
import org.omwg.logicalexpression.Constants;
import org.omwg.logicalexpression.LogicalExpression;
import org.omwg.logicalexpression.MembershipMolecule;
import org.omwg.logicalexpression.Molecule;
import org.omwg.logicalexpression.terms.Term;
import org.omwg.ontology.Attribute;
import org.omwg.ontology.Concept;
import org.omwg.ontology.DataValue;
import org.omwg.ontology.OntologyElement;
import org.omwg.ontology.Variable;
import org.omwg.ontology.WsmlDataType;
import org.wsmo.common.IRI;
import org.wsmo.common.Identifier;
import org.wsmo.factory.*;
import org.wsmostudio.runtime.WSMORuntime;

public class WsmoLogExprExport
{
	private Set<ModelWithConnections> visitedModels = new HashSet<ModelWithConnections>();
	private Set<OntologyShell> usedOntologies = new HashSet<OntologyShell>();
	private Map<String, Variable> usedVariables = new HashMap<String, Variable>();
	private Set<Variable> existsVariables = new HashSet<Variable>();

	public static final LogicalExpressionFactory leFactory = 
        WSMORuntime.getRuntime().getLogExprFactory();
    public static final DataFactory dataFactory = 
        WSMORuntime.getRuntime().getDataFactory();
	
	private static final WsmlDataType integerDataType = dataFactory.createWsmlDataType(WsmlDataType.WSML_INTEGER);
	private static final WsmlDataType decimalDataType = dataFactory.createWsmlDataType(WsmlDataType.WSML_DECIMAL);
	private static final WsmlDataType doubleDataType = dataFactory.createWsmlDataType(WsmlDataType.WSML_DOUBLE);
	private static final WsmlDataType floatDataType = dataFactory.createWsmlDataType(WsmlDataType.WSML_FLOAT);
	private static final WsmlDataType booleanDataType = dataFactory.createWsmlDataType(WsmlDataType.WSML_BOOLEAN);

	private Variable getVariable(String varName) {
		return this.usedVariables.get(varName);
	}

	private void addUsedOntology(OntologyElement element) {
		OntologyShell ontology = AxiomEditorPlugin.ontoStore.getOntologyContaining(element);
		if (ontology != null && !this.usedOntologies.contains(element) &&
				!Util.isWsmlSyntaxOntology(ontology.getOntology())) {
			this.usedOntologies.add(ontology);
		}
	}

	/**
	 * @return The expressions corresponding to all not visited models.
	 */
	private List<LogicalExpression> getExpressionsList(List<? extends ModelWithConnections> models) {
		List<LogicalExpression> expressions = new ArrayList<LogicalExpression>();
		for (ModelWithConnections model: models) {
			LogicalExpression le = exportModelWithConnections(model);
			if (le != null) {
				expressions.add(le);
			}
		}
		return expressions;
	}

	/**
	 *	"simple head" and "complex tail"
	 * e.g.
	 *	?trip memberOf boatTrip [ ... ]
	 *	and
	 *	(	?start = innsbruckHbf
	 *		or ?start = stuttgartHbf
	 *  )
	 *	and
	 *	(	not ?end memberOf station [ ... ]
	 *	}
	 * NOTE: The list of operands is destroyed.
	 */
	private LogicalExpression createNaryOperator(int operatorType, List expressions) {
		if (expressions == null || expressions.isEmpty()) {
			return null;
		}
		LogicalExpression head = (LogicalExpression)expressions.get(0);
		expressions.remove(0);
		LogicalExpression tail = createNaryOperator(operatorType, expressions);
		if (head == null) {
			return tail;
		}
		if (tail == null) {
			return head;
		}
		if (operatorType == OperatorModel.OPERATOR_AND) {
			return leFactory.createConjunction(head, tail);
		}
		else if (operatorType == OperatorModel.OPERATOR_OR) {
			return leFactory.createDisjunction(head, tail);
		}
		else {
			throw new IllegalArgumentException("Illegal N-ary operator type: " + operatorType);
		}
	}

	private LogicalExpression appendRefiningModels(LogicalExpression declaration, List<ModelWithConnections> refiningModels) {
		List<LogicalExpression> expressions = getExpressionsList(refiningModels);
		// Return a conjunction with the declaration as the first operand.
		expressions.add(0, declaration);
		return createNaryOperator(OperatorModel.OPERATOR_AND, expressions);
	}

	/**
	 *	areSomehowRelated(?BuyerContact, _#, ?CarRental) and
	 *	(    ?BuyerContact = myContract
	 *	) 
	 *	and
	 *	(    ?CarRental memberOf CarRentalBookingPreferences [ .. ]
	 *	)
	 */
	private LogicalExpression exportRelationModel(RelationModel relationModel) {
		addUsedOntology(relationModel.getRelation());

		// Compose the relation declaration.
		Identifier relationName = relationModel.getRelation().getIdentifier();
		List<Term> paramTerms = getRelationParametersList(relationModel);
		Atom declaration = leFactory.createAtom(relationName, paramTerms);

		// Append the expressions for all not visited refining models.
		List<ModelWithConnections> refiningModels = relationModel.getRefinableList().listRefiningModels();
		return appendRefiningModels(declaration, refiningModels); 
	}

	/**
	 *	?BuyerContact, _#, ?CarRental
	 */
	private List<Term> getRelationParametersList(RelationModel relation) {
		List<Term> params = new ArrayList<Term>();
		Iterator i = relation.getRefinableList().getChildren().iterator();
		while (i.hasNext()) {
			ParameterModel paramModel = (ParameterModel)i.next();
			if (paramModel.getVarNameModel() != null) {
				String varName = paramModel.getVarNameModel().getName(); 
				params.add(getVariable(varName));
			}
			else {
				params.add(WSMORuntime.getRuntime().getWsmoFactory().createAnonymousID());
			}
		}
		return params;
	}

	/**	?PickupLocation memberOf loc#location
	 *	[	contactInformation hasValue ?BuyerContact,
	 *		bookingPreferences hasValue ?CarRental
	 *	]
	 */	
	private LogicalExpression exportVariableModel(VariableModel variableModel) {
		addUsedOntology(variableModel.getConcept());

		// Compose the variable declaration.
		String varName = variableModel.getVarNameModel().getName();
		Identifier conceptId = variableModel.getConcept().getIdentifier();
		MembershipMolecule membership = leFactory.createMemberShipMolecule(
				getVariable(varName), conceptId);

		LogicalExpression declaration = membership; 
		// Append the attribute declarations.
		List<Molecule> molecules = new ArrayList<Molecule>(
				getVariableAttributesList(variableModel));
		// Prepend the membership.
		if (!molecules.isEmpty()) {
			molecules.add(0, membership);
			declaration = leFactory.createCompoundMolecule(molecules);
		}
		
		// Append the expressions for all not visited refining models.
		List<ModelWithConnections> refiningModels = variableModel.getRefinableList().listRefiningModels();
		return appendRefiningModels(declaration, refiningModels); 
	}

	/**
	 *	[	contactInformation hasValue {?BuyerContact},
	 *		bookingPreferences hasValue ?CarRental
	 *	]
	 */
	private List<AttributeValueMolecule> getVariableAttributesList(VariableModel variableModel) {
		List<AttributeValueMolecule> attrs = new ArrayList<AttributeValueMolecule>();
		String varName = variableModel.getVarNameModel().getName();
		Variable var = getVariable(varName);

		Iterator i = variableModel.getRefinableList().getChildren().iterator();
		while (i.hasNext()) {
			AttributeModel attrModel = (AttributeModel)i.next();
			if (attrModel.getVarNameModel() != null) {
				Attribute attr = ((AttributeShell)attrModel.getRefinable()).getAttribute();
				Identifier attrId = attr.getIdentifier();
				String attrVarName = attrModel.getVarNameModel().getName();
				attrs.add(leFactory.createAttributeValue(var, attrId, getVariable(attrVarName)));
			}
		}
		return attrs;
	}

	/**
	 *	?Human1 = Susan 
	 */
	private Atom createEquality(VarNameHolderModel model, Term value) {
		List<Term> compare = new ArrayList<Term>();
		compare.add(getVariable(model.getVarNameModel().getName()));
		compare.add(value);
		IRI eq = WSMORuntime.getRuntime().getWsmoFactory().createIRI(Constants.EQUAL);
		return leFactory.createAtom(eq, compare);
	}

	private LogicalExpression exportInstanceModel(InstanceModel instance) {
		addUsedOntology(instance.getInstance());

		Identifier value = instance.getInstance().getIdentifier();
		return createEquality(instance, value);
	}

	/**
	 *	?weight memberOf XMLSchema#boolean and
	 *	?weight = XMLSchema#boolean("true")
	 */
	private LogicalExpression exportValueInstanceModel(ValueInstanceModel valueInstance) {
		addUsedOntology(valueInstance.getConcept());

		// Create the membership declaration.
		Concept concept = valueInstance.getConcept();
		IRI conceptIri = (IRI)concept.getIdentifier();
		String varName = valueInstance.getVarNameModel().getName();
		MembershipMolecule membership = leFactory.createMemberShipMolecule(
				getVariable(varName), conceptIri);

		// If no value is entered do not add an equality atom but return the membership only.
		String valueText = valueInstance.getValue(); 
		if (valueText == null) {
			return membership;
		}

		// Create the value.
		DataValue value = null;
		try {
			DataFactory df = WSMORuntime.getRuntime().getDataFactory();
			if (Util.isSubConcept(concept, integerDataType)) {
				value = df.createWsmlInteger(valueText);
			} 
			else if (Util.isSubConcept(concept, decimalDataType)) {
				value = df.createWsmlDecimal(valueText);
			} 
			else if (Util.isSubConcept(concept, doubleDataType)) {
				value = df.createWsmlDouble(valueText);
			} 
			else if (Util.isSubConcept(concept, floatDataType)) {
				value = df.createWsmlFloat(valueText);
			} 
			else if (Util.isSubConcept(concept, booleanDataType)) {
				value = df.createWsmlBoolean(valueText);
			} 
			else {
				value = df.createWsmlString(valueText);
			}
		}
		catch (IllegalArgumentException ex) {
			System.out.println("Failed creating " + conceptIri + ": " + ex.getMessage());
			return membership;
		}

		// Append the equality. 
		Atom equality = createEquality(valueInstance, value);
		return leFactory.createConjunction(membership, equality);
	}

	/**
	 *	a) not (...)
	 *	b) (...) and ((...) and (...))
	 */
	private LogicalExpression exportOperatorModel(OperatorModel operator) {
		List<ModelWithConnections> operands = operator.listOperands();
		List<LogicalExpression> expressions = getExpressionsList(operands);
		if (expressions.isEmpty()) {
			return null;
		}
		if (operator.getOperatorType() == OperatorModel.OPERATOR_NOT) {
			// Export unary operator model.
			return leFactory.createNegation(expressions.get(0));
		}
		// Export N-ary operator model.
		return createNaryOperator(operator.getOperatorType(), expressions);
	}

	/**
	 *	?PickupLocation memberOf loc#location
	 *	[	...
	 *	] and
	 *	?ReturnLocation memberOf loc#location and
	 *	?DoorCount memberOf xsd#integer and
	 */
	private LogicalExpression exportAxiomRootModel(AxiomRootModel axiomRoot) {
		ModelWithConnections top = axiomRoot.getTopModel();
		if (top != null) {
    		return exportModelWithConnections(top);
		}
		return null;
	}

	private LogicalExpression exportMemberOfModel(MemberOfModel model) {
		LogicalExpression memberOfExpr = null;
		if (model instanceof VariableModel) { 
			memberOfExpr = exportVariableModel((VariableModel)model);
		}
		else if (model instanceof InstanceModel) { 
			memberOfExpr = exportInstanceModel((InstanceModel)model);
		}
		else if (model instanceof ValueInstanceModel) { 
			memberOfExpr = exportValueInstanceModel((ValueInstanceModel)model);
		}
		// Attach the floating relations which have a parameter connected to this model.
		List<LogicalExpression> expressions = getExpressionsList(model.listChildRelations());

		// Return a conjunction with the very memberOf expression as the first operand.
		expressions.add(0, memberOfExpr);
		return createNaryOperator(OperatorModel.OPERATOR_AND, expressions);
	}
	
	private LogicalExpression exportModelWithConnections(ModelWithConnections model) {
		// Prevent the same model from appearing more than once in the expression.
		if (this.visitedModels.contains(model)) {
			return null;
		}
		this.visitedModels.add(model);
		
		if (model instanceof MemberOfModel) {
			return exportMemberOfModel((MemberOfModel)model);
		}
		if (model instanceof RelationModel) { 
			return exportRelationModel((RelationModel)model);
		}
		if (model instanceof OperatorModel) { 
			return exportOperatorModel((OperatorModel)model);
		}
    	return null;
	}

	/**
	 * Creates a Variable object for every VarNameModel used in the axiom.
	 * This simplifies the access to variables during expression construction.
	 */
	private void prepareVariables(VarNameList varNameList) {
		this.usedVariables.clear();
		this.existsVariables.clear();
		Iterator i = varNameList.iterator();
		while (i.hasNext()) {
			VarNameModel varNameModel = (VarNameModel)i.next();
			String varName = varNameModel.getName();
			Variable var = WSMORuntime.getRuntime().getWsmoFactory().createVariable(varName);
			this.usedVariables.put(varName, var);
			if (true == varNameModel.getExistsFlag()) {
				this.existsVariables.add(var);
			}
		}
	}

	/**
	 * Converts an axiom model to a WSMO Logical expression.
	 */
	public LogicalExpression createLogExpr(AxiomContentsModel contentsModel) {
		this.visitedModels.clear();
		this.usedOntologies.clear();
		prepareVariables(contentsModel.getVarNameList());

		AxiomRootModel root = contentsModel.getAxiomRoot();
		LogicalExpression le = exportAxiomRootModel(root);
		if (!this.existsVariables.isEmpty()) {
			le = leFactory.createExistentialQuantification(this.existsVariables, le);
		}
		return le;
	}

	/**
	 * @return All ontologies referenced by the last axiom model which was passed to createLogExpr().
	 * NOTE: createLogExpr() must be called first.
	 */
	public Set<OntologyShell> listUsedOntologies() {
		return this.usedOntologies;
	}

}
