package org.infrawebs.axiom.logexpr;

import java.util.*;

import org.infrawebs.ontology.shells.OntologyShell;
import org.omwg.logicalexpression.LogicalExpression;
import org.omwg.ontology.*;
import org.wsmo.common.*;
import org.wsmo.common.exception.InvalidModelException;
import org.wsmo.factory.WsmoFactory;
import org.wsmo.mediator.Mediator;
import org.wsmo.service.*;
import org.wsmostudio.runtime.WSMORuntime;

public class WsmoTopEntityExport
{
	public static final String DEFAULT_NAMESPACE = "http://www.infrawebs.org/axiom#";
	private static final WsmoFactory wsmoFactory = 
        WSMORuntime.getRuntime().getWsmoFactory();

	/**
	 *	namespace {_"http://www.infrawebs.org/axiom#",
	 *		dc	_"http://purl.org/dc/elements/1.1#",
	 *		XMLSchema	_"http://www.w3.org/2001/XMLSchema#",
	 *		example	_"http://www.example.org/ontologies/example#"
	 *	}
	 *	...
	 *	importsOntology {
	 *		 _"http://www.example.org/ontologies/example"
	 *	}
	 */
	private void addNamespaces(TopEntity topEntity, Set<OntologyShell> usedOntologies) {
		// Always add the Dublin Core namespace because the axiom uses the NFP dc#description.
		Namespace dublinCore = wsmoFactory.createNamespace("dc",
				wsmoFactory.createIRI("http://purl.org/dc/elements/1.1"));
		topEntity.addNamespace(dublinCore);

		for (OntologyShell ontology: usedOntologies) {
			IRI iri = (IRI)ontology.getOntology().getIdentifier();
			Namespace ns = wsmoFactory.createNamespace(ontology.getPrefix(), iri);
			topEntity.addNamespace(ns);
			topEntity.addOntology(ontology.getOntology());
		}
	}

	private Axiom createEmptyAxiom(String axiomName) {
		IRI iri = wsmoFactory.createIRI(DEFAULT_NAMESPACE + axiomName);
		Axiom axiom = wsmoFactory.createAxiom(iri);
		
		// Clear any old lists because the web service might have been cached by the WSMO factory.
		Iterator i = axiom.listDefinitions().iterator();
		while (i.hasNext()) {
			axiom.removeDefinition((LogicalExpression)i.next());
			i = axiom.listDefinitions().iterator();
		}

		return axiom;
	}

	/**
	 *	nonFunctionalProperties
	 *		dc#description hasValue "The booking ticket is delivered to the buyer via email"
	 *	endNonFunctionalProperties
	 * NOTE: Assumes that nfps has String keys and String values. 
	 */
	private static void addNFPs(Axiom axiom, Map nfps) {
		Iterator i = nfps.keySet().iterator();
		while (i.hasNext()) {
			String key = (String)i.next();
			String strValue = (String)nfps.get(key);
			IRI iri = wsmoFactory.createIRI(key);
			Value value = WSMORuntime.getRuntime().getDataFactory().createWsmlString(strValue);
			try {
				axiom.addNFPValue(iri, value);
			}
			catch (InvalidModelException ex) {
				System.out.println("Error adding NFP to axiom: " + ex.getMessage());
			}
		}
	}

	public Axiom createAxiom(LogicalExpression le, String axiomName, Map nfps) {
		Axiom axiom = createEmptyAxiom(axiomName);
		addNFPs(axiom, nfps);
		// Add the new definition.
		if (le != null) {
			axiom.addDefinition(le);
		}
		return axiom;
	}

	/**
	 * Clears any old lists because the entity might have been cached by the WSMO factory.
	 */
	private void clearTopEntity(TopEntity top) {
		Iterator i = top.listNamespaces().iterator();
		while (i.hasNext()) {
			top.removeNamespace((Namespace)i.next());
		}
		i = top.listOntologies().iterator();
		while (i.hasNext()) {
			top.removeOntology((Ontology)i.next());
		}
		i = top.listMediators().iterator();
		while (i.hasNext()) {
			top.removeMediator((IRI)i.next());
		}
		try {
			i = top.listNFPValues().keySet().iterator();
			while (i.hasNext()) {
				top.removeNFP((IRI)i.next());
			}
		}
		catch (InvalidModelException ex) {
			System.out.println("Error clearing top entity: " + ex.getMessage());
		}
	}

	/**
	 * Clears any old lists because the entity might have been cached by the WSMO factory.
	 */
	private void clearCapability(Capability capability) {
		clearTopEntity(capability);
		try {
			Iterator i = capability.listPreConditions().iterator();
			while (i.hasNext()) {
				capability.removePreCondition((Axiom)i.next());
			}
			i = capability.listAssumptions().iterator();
			while (i.hasNext()) {
				capability.removeAssumption((Axiom)i.next());
			}
			i = capability.listPostConditions().iterator();
			while (i.hasNext()) {
				capability.removePostCondition((Axiom)i.next());
			}
			i = capability.listEffects().iterator();
			while (i.hasNext()) {
				capability.removeEffect((Axiom)i.next());
			}
			i = capability.listSharedVariables().iterator();
			while (i.hasNext()) {
				capability.removeSharedVariable((Variable)i.next());
			}
		}
		catch (InvalidModelException ex) {
			System.out.println("Error clearing capability: " + ex.getMessage());
		}
	}

	private WebService createEmptyWebService() {
		// Create the default namespace.
		Namespace defaultNs = wsmoFactory.createNamespace(
				"", wsmoFactory.createIRI(DEFAULT_NAMESPACE));

		IRI wsIri = wsmoFactory.createIRI(defaultNs, "dummyWebService");
		WebService webService = wsmoFactory.createWebService(wsIri);
		clearTopEntity(webService);
		webService.setDefaultNamespace(defaultNs);

		IRI capIri = wsmoFactory.createIRI(defaultNs, "dummyCapability");
		Capability capability = wsmoFactory.createCapability(capIri);
		clearCapability(capability);
		capability.setDefaultNamespace(defaultNs);
		webService.setCapability(capability);

		return webService;
	}

	public WebService createWebService(Set<OntologyShell> usedOntologies, Axiom axiom) {
		WebService webService = createEmptyWebService();
		if (usedOntologies != null) {
			addNamespaces(webService, usedOntologies);
		}
		if (axiom != null) {
			// Set the axiom as the web service's only precondition.
			try {
				webService.getCapability().addPreCondition(axiom);
			}
			catch (InvalidModelException ex) {
				System.out.println("Error adding axiom to precondition: " + ex.getMessage());
			}
		}
		return webService;
	}

}
