package org.infrawebs.axiom.model;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.ui.views.properties.IPropertyDescriptor;
import org.eclipse.ui.views.properties.PropertyDescriptor;
import org.infrawebs.axiom.plugin.AxiomEditorPlugin;
import org.infrawebs.common.Util;
import org.infrawebs.ontology.RestorationException;
import org.omwg.ontology.Concept;
import org.wsmo.common.Identifier;

public class MemberOfModel extends VarNameHolderModel implements Compactable
{
	private static final long serialVersionUID = 200512242100L;
	
	public static final String ID_PROPERTY_TYPE = "ID_PROPERTY_TYPE";

	private transient Concept concept = null;	
	private boolean isCompact = false;
	private Dimension expandedWidth = null;

	public MemberOfModel(AxiomContentsModel contents, Concept concept) {
		super(contents);
		this.concept = concept;
		String prefix = Util.getLocalName(concept);
		setVarNameModel(new VarNameModel(contents.getVarNameList(), prefix));
	}

	public Concept getConcept() {
		return this.concept;
	}
	
	@Override
	public IPropertyDescriptor[] createPropertyDescriptors() {
		IPropertyDescriptor[] superDescriptors = super.createPropertyDescriptors();
		int oldLen = superDescriptors.length;
		IPropertyDescriptor[] descriptors = new IPropertyDescriptor[oldLen + 1];
		System.arraycopy(superDescriptors, 0, descriptors, 0, oldLen);
		descriptors[oldLen + 0] = new PropertyDescriptor(ID_PROPERTY_TYPE, "Type");
		return descriptors;
	}

	@Override
	public Object getPropertyValue(Object propName) {
		if (propName.equals(ID_PROPERTY_TYPE)) {
			return AxiomEditorPlugin.ontoStore.getPrefixedName(this.concept);
		} 
		return super.getPropertyValue(propName);
	}

	@Override
	public void setConstraint(Object constraint) {
		Rectangle newRect = (Rectangle)constraint;
		if (!isCompact()) {
			// Keep models auto sized.
			newRect.height = -1;
		}
		else {
			// Keep original compact size.
			newRect.setSize(COMPACT_SIZE);			
		}
		super.setConstraint(newRect);
	}

	public boolean isCompact() {
		return this.isCompact;
	}

	public void setCompact(boolean isCompact) {
		boolean oldIsCompact = this.isCompact; 
		this.isCompact = isCompact;

		this.expandedWidth = recalculateConstraint(
				this, this.isCompact, this.expandedWidth);
		
		firePropertyChange(ID_PROPERTY_COMPACT,
				new Boolean(oldIsCompact), new Boolean(isCompact));
	}
	
	/**
	 * Called by setCompact() to find the new constraint of the model.
	 * TODO: This method was made static so that RelationModels can use it
	 * (to avoid duplicate code). It should be redesigned.
	 */
	protected static Dimension recalculateConstraint(
			ModelWithConnections model, boolean isCompact, Dimension expandedWidth) {
		Rectangle oldRect = (Rectangle)model.getConstraint();
		if (oldRect == null) {
			return null;
		}

		Dimension newSize = null;
		Dimension result = null;
		if (isCompact) {
			newSize = COMPACT_SIZE;
			// Return the width so that it can be restored.
			result = oldRect.getSize();
		}
		else {
			// Use a default size in case the model has never been expanded.
			newSize = (expandedWidth != null) ? expandedWidth : new Dimension(100, -1);
			result = null;
		}
		Point topLeft = oldRect.getCenter().getTranslated(
				-newSize.width / 2, -newSize.height / 2);
		model.setConstraint(new Rectangle(topLeft, newSize));
		return result;
	}

	/**
	 * Finds relations which should appear as children of this model in the expression tree.
	 * Such relations are not fixed in the expression but are floating. They are attached
	 * after one of the models that they relate - the deepest one (the first at that depth). 
	 * @see RelationModel.isFixedInExpression()
	 * @return A list of relations for which this model is their deepest argument.
	 */
	public List<RelationModel> listChildRelations() {
		List<RelationModel> result = new ArrayList<RelationModel>();
		List<ModelWithConnections> parents = listGrafParents();
		for (int i = 0; i < parents.size(); i++) {
			if (parents.get(i) instanceof ParameterModel) {
				ParameterModel param = (ParameterModel)parents.get(i);
				ParameterListModel paramList = (ParameterListModel)param.getOwner();
				RelationModel relation = (RelationModel)paramList.getOwner();
				if (!relation.isFixedInExpression()) {
					ParameterModel deepest = paramList.getDeepestParameter();
					if (deepest != null && deepest.getRefiningModel() == this) {
						result.add(relation);
					}
				}
			}
		}
		return result; 
	}

	private void writeObject(java.io.ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
		// Write just the concept identifier, not the concept itself.
		out.writeObject(this.concept.getIdentifier());
	}

	private void readObject(java.io.ObjectInputStream in)
			throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		// Read the concept identifier.
		Identifier conceptId = (Identifier)in.readObject();
		
		// Restore the definition of the concept.
		this.concept = (Concept)AxiomEditorPlugin.ontoStore.findDefinition(
				conceptId, Concept.class);
		if (this.concept == null) {
			throw new RestorationException("Concept " + conceptId + " not found.");
		}
	}

}
