package org.infrawebs.axiom.model;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.ui.views.properties.IPropertyDescriptor;
import org.eclipse.ui.views.properties.PropertyDescriptor;
import org.infrawebs.axiom.model.commands.VarNameChangeCommand;
import org.infrawebs.axiom.operations.ModelOperations;
import org.infrawebs.axiom.plugin.AxiomEditorPlugin;
import org.infrawebs.common.Util;
import org.infrawebs.ontology.shells.RefinableShell;

public abstract class RefinableModel extends VarNameHolderModel
{
	private static final long serialVersionUID = 200512242100L;

	public static final String ID_PROPERTY_NAME = "ID_PROPERTY_NAME";
	public static final String ID_PROPERTY_TYPE = "ID_PROPERTY_TYPE";
	public static final String ID_PROPERTY_CONSTRAINING = "ID_PROPERTY_CONSTRAINING";

	protected transient RefinableShell refinable = null;
	private RefinableListModel owner = null;

	public RefinableModel(AxiomContentsModel contents, RefinableShell refinable) {
		super(contents);
		this.refinable = refinable;
	}

	public RefinableListModel getOwner() {
		return this.owner;
	}

	// This setter is called on owner.addChild(<this object>).
	void setOwner(RefinableListModel owner) {
		this.owner = owner;
	}
	
	public RefinableShell getRefinable() {
		return this.refinable;
	}
	
	public abstract VarNameModel createVarNameModel();
	
	@Override
	public IPropertyDescriptor[] createPropertyDescriptors() {
		IPropertyDescriptor[] superDescriptors = super.createPropertyDescriptors();
		int oldLen = superDescriptors.length;
		IPropertyDescriptor[] descriptors = new IPropertyDescriptor[oldLen + 3];
		System.arraycopy(superDescriptors, 0, descriptors, 0, oldLen);
		descriptors[oldLen + 0] = new PropertyDescriptor(ID_PROPERTY_NAME, "Name");
		descriptors[oldLen + 1] = new PropertyDescriptor(ID_PROPERTY_TYPE, "Types");
		descriptors[oldLen + 2] = new PropertyDescriptor(ID_PROPERTY_CONSTRAINING, "Constraining");
		return descriptors;
	}
	
	@Override
	public Object getPropertyValue(Object id) {
		if (id.equals(ID_PROPERTY_NAME)) {
			return Util.getLocalName(this.refinable);
		}
		else if (id.equals(ID_PROPERTY_TYPE)) {
			return AxiomEditorPlugin.ontoStore.getPrefixedNames(
					this.refinable.listTypes());
		}
		else if (id.equals(ID_PROPERTY_CONSTRAINING)) {
			return this.refinable.isConstraining() ? "true" : "false";
		}
		else {
			return super.getPropertyValue(id);
		}
	}
	
	/**
	 * Changes the variable name or if this is not possible removes/restores the whole varNameModel.
	 */
	private void specialChangeVariable(String varName) {
		boolean isEmptyName = (varName == null || varName.length() == 0);
		//CommandStack stack = AxiomEditorPlugin.getAxiomEditor().getEditDomain().getCommandStack();

		if (this.getVarNameModel() != null) {
			if (!isEmptyName) {
				// Do a regular renaming of variables.
				super.setPropertyValue(VarNameModel.ID_PROPERTY_VARNAME, varName);
			}
			else {
				// If the user deleted the variable name, detach the variable.
				// TODO: Another command appears in the stack after this command. It must not be executed.  
				ModelOperations.executeCommand(new VarNameChangeCommand(this, null));
			}
		}
		else {  // varNameModel is null
			if (!isEmptyName) {
				// If the user manually entered a variable name, create and attach a new variable.
				VarNameModel varNameModel = new VarNameModel(this.contents.getVarNameList(), varName);
				// TODO: Another command appears in the stack after this command. It must not be executed.  
				ModelOperations.executeCommand(new VarNameChangeCommand(this, varNameModel));
			}
			else {
				// Setting an empty name to a missing variable. Nothing to do.
			}
		}
	}

	@Override
	public void setPropertyValue(Object id, Object value) {
		if (id.equals(VarNameModel.ID_PROPERTY_VARNAME)) {
			specialChangeVariable((String)value);
		}
		else {
			super.setPropertyValue(id, value);
		}
	}

	/**
	 * Refinables return their owner variable or relation
	 * (jumping over their direct owner which is just a container). 
	 */
	@Override
	public List<ModelWithConnections> listGrafParents() {
		List<ModelWithConnections> parents = new ArrayList<ModelWithConnections>();
		parents.add(getOwner().getOwner());
		return parents;
	}

	/**
	 * A RefinableModel should be refined by exactly 1 subexpression.
	 */  
	public ModelWithConnections getRefiningModel() {
		List<ModelWithConnections> children = listGrafChildren();
		if (children.size() > 0) {
			return children.get(0); 
		}
		return null;
	}
	
	/**
	 * Ends the recursive search for the connection chain origins.
	 * RefinableModels are origins themselves.
	 */
	@Override
	public List<RefinableModel> listChainOrigins() {
		List<RefinableModel> origins = new ArrayList<RefinableModel>();
		origins.add(this);
		return origins;
	}

}
