package org.infrawebs.axiom.text;

import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.infrawebs.axiom.model.AttributeModel;
import org.infrawebs.axiom.model.AxiomContentsModel;
import org.infrawebs.axiom.model.AxiomRootModel;
import org.infrawebs.axiom.model.InstanceModel;
import org.infrawebs.axiom.model.MemberOfModel;
import org.infrawebs.axiom.model.ModelWithConnections;
import org.infrawebs.axiom.model.OperatorModel;
import org.infrawebs.axiom.model.ParameterModel;
import org.infrawebs.axiom.model.RelationModel;
import org.infrawebs.axiom.model.ValueInstanceModel;
import org.infrawebs.axiom.model.VarNameList;
import org.infrawebs.axiom.model.VarNameModel;
import org.infrawebs.axiom.model.VariableModel;
import org.infrawebs.axiom.plugin.AxiomEditorPlugin;
import org.infrawebs.common.Util;
import org.infrawebs.ontology.shells.AttributeShell;
import org.infrawebs.ontology.shells.OntologyShell;
import org.omwg.logicalexpression.Constants;
import org.omwg.ontology.Attribute;
import org.omwg.ontology.Concept;
import org.omwg.ontology.OntologyElement;

public class AxiomTextGenerator
{
	private Set<ModelWithConnections> visitedModels = new HashSet<ModelWithConnections>();
	private Set<OntologyShell> usedOntologies = new HashSet<OntologyShell>();
	private AxiomContentsModel contentsModel = null;

	public AxiomTextGenerator(AxiomContentsModel contentsModel) {
		this.contentsModel = contentsModel;
	}

	private static String indented(String text) {
		return indented(text, 1);
	}

	private static String indented(String text, int tabCount) {
		String[] stringArray = text.split("\n");
		String result = "";
		String offsetString = "";

		for (int i = 0; i < tabCount; i++)    	
			offsetString += "\t";
    	
		for (int i = 0; i < stringArray.length; i++) {
			result += offsetString + stringArray[i];
			// do not put \n on the last line
			if (i < stringArray.length - 1)
				result += "\n";
		}

		return result;
	}

	private void addUsedOntology(OntologyElement element) {
		OntologyShell ontology = AxiomEditorPlugin.ontoStore.getOntologyContaining(element);
		if (ontology != null && !this.usedOntologies.contains(element) &&
				!Util.isWsmlSyntaxOntology(ontology.getOntology())) {
			this.usedOntologies.add(ontology);
		}
	}

	//	areSomehowRelated(?BuyerContact, _#, ?CarRental) and
	//	(    ?BuyerContact = myContract
	//	) 
	//	and
	//	(    ?CarRental memberOf CarRentalBookingPreferences [ .. ]
	//	)
	private String getRelationModelText(RelationModel relation) {
		addUsedOntology(relation.getRelation());
		
		String remationName = AxiomEditorPlugin.ontoStore.getPrefixedName(relation.getRelation());
		List refinements = relation.getRefinableList().listRefiningModels();
		return remationName + "(" + getRelationParameterListText(relation) + ")"
				+ getRefinementsText(refinements);
	}

	//	?BuyerContact, _#, ?CarRental
	private String getRelationParameterListText(RelationModel relation) {
		List paramList = relation.getRefinableList().getChildren();
		String paramText = "";
		for (int i = 0; i < paramList.size(); i++) {
			ParameterModel param = (ParameterModel)paramList.get(i);
			paramText += getParameterModelText(param);
    		// Do not put a comma after the last parameter.
    		if (i < paramList.size() - 1) {
    			paramText += ", ";
    		}
		}
		return paramText;
	}

	//	?BuyerContact OR _#
	private static String getParameterModelText(ParameterModel parameter) {
		// If the parameter is refined, add its variable name. Otherwise output an anonymous variable.
		ModelWithConnections refinement = parameter.getRefiningModel();
		if (refinement != null) {
			return Constants.VARIABLE_DEL + parameter.getVarNameModel().getName();
		}
		return Constants.ANONYMOUS_ID_NOTATION;				
	}

	//	?PickupLocation memberOf loc#location
	//	[	contactInformation hasValue ?BuyerContact,
	//		bookingPreferences hasValue ?CarRental
	//	]	
	private String getVariableModelText(VariableModel variable) {
		addUsedOntology(variable.getConcept());
		
		String type = AxiomEditorPlugin.ontoStore.getPrefixedName(variable.getConcept());
		List refinements = variable.getRefinableList().listRefiningModels();
		return Constants.VARIABLE_DEL + variable.getVarNameModel().getName() + 
				" memberOf " + type + 
				getVariableAttributesDeclarationText(variable) + 
				getRefinementsText(refinements);
	}

	//	[	contactInformation hasValue {?BuyerContact},
	//		bookingPreferences hasValue ?CarRental
	//	]	
	private static String getVariableAttributesDeclarationText(VariableModel variable) {
		List attrList = variable.getRefinableList().listRefinedAttributes();
		String text = "";
		for (int i = 0; i < attrList.size(); i++) {
			AttributeModel attrModel = (AttributeModel)attrList.get(i);
			text += Util.getLocalName(attrModel.getRefinable()) + " hasValue ";

			String varName = attrModel.getVarNameModel().getName();
			Attribute attr = ((AttributeShell)attrModel.getRefinable()).getAttribute();
			text += (attr.getMaxCardinality() <= 1)
					? Constants.VARIABLE_DEL + varName
					: "{" + Constants.VARIABLE_DEL + varName + "}";

			// Do not put a comma after the last attribute.
    		if (i < attrList.size() - 1) {
    			text += ",\n";
    		}
		}
		return (text.length() > 0) ? "\n[" + indented(text) + "\n]" : "";
	}

	//	and
	//	(    ?start = innsbruckHbf
	//	) 
	//	and
	//	(    ?end memberOf station [ .. ]
	//	)
	private String getRefinementsText(List refinements) {
		String text = "";
		for (int i = 0; i < refinements.size(); i++) {
			ModelWithConnections refinement = (ModelWithConnections)refinements.get(i);
			String refinementText = getModelWithConnectionText(refinement);
			if (refinementText.length() > 0) {
				text += " and\n(" + indented(refinementText) + "\n)";
			}
		}
		return text;
	}

	//	and
	//	example#sameZodiac(?Human1, ?Human2)
	private String getChildRelationsText(List relations) {
		String text = "";
		for (int i = 0; i < relations.size(); i++) {
			RelationModel relation = (RelationModel)relations.get(i);
			String relationText = getModelWithConnectionText(relation);
			if (relationText.length() > 0) {
				text += " and\n" + relationText + "\n";
			}
		}
		return text;
	}

	// ?Human1 = Susan
	private String getInstanceModelText(InstanceModel constant) {
		addUsedOntology(constant.getConcept());
		
		return Constants.VARIABLE_DEL + constant.getVarNameModel().getName() 
				+ " " + Constants.EQUAL_NOTATION + " " + constant.getValue();
	}

	// decimal: 123
	// string: "test"
	// boolean: _boolean("true")
	private String formatValueLiteral(Concept concept, String valueText) {
		// Use WSML shortcuts if possible.
		if (Util.isWsmlDataType(concept, "decimal") || Util.isWsmlDataType(concept, "integer")) {
			return valueText;
		}
		if (Util.isWsmlDataType(concept, "string")) {
			return Constants.STRING_DEL_START + valueText + Constants.STRING_DEL_END;
		}
		if (Util.isWsmlDataType(concept, "iri")) {
			return Constants.IRI_DEL_START + valueText + Constants.IRI_DEL_END;
		}
		String typeName = AxiomEditorPlugin.ontoStore.getPrefixedName(concept);
		return typeName + "(" + valueText + ")";
	}

	// ?weight memberOf XMLSchema#boolean and
	// ?weight = XMLSchema#boolean("true")
	private String getValueInstanceText(ValueInstanceModel valueInstance) {
		Concept concept = valueInstance.getConcept();
		addUsedOntology(concept);

		String typeName = AxiomEditorPlugin.ontoStore.getPrefixedName(concept);
		String varName = Constants.VARIABLE_DEL + valueInstance.getVarNameModel().getName(); 
		String text = varName + " memberOf " + typeName;
		String valueText = valueInstance.getValue(); 
		if (valueText != null) {
			text += " and\n" + varName + " " + Constants.EQUAL_NOTATION + " " +
				formatValueLiteral(concept, valueText);
		}
		return text;
	}
	
	private String getOperatorInfixModelText(OperatorModel operator) {
		// Optionally, operators with insufficient operands may be discarded.
//		if (model.getModelSourceConnections().size() < 2) {
//			return "";
//		}

		String text = "";
		List operands = operator.listOperands();
		for (int i = 0; i < operands.size(); i++) {
			ModelWithConnections operand = (ModelWithConnections)operands.get(i);
			text += "(" + indented(getModelWithConnectionText(operand)) + "\n)"; 
    		// Do not put the infix operator after the last operand.
    		if (i < operands.size() - 1) {
    			text += "\n" + operator.getOperatorTypeAsString().toLowerCase() + "\n";
    		}
		}
		return text;
	}

	private String getOperatorPrefixModelText(OperatorModel operator) {
		List operands = operator.listOperands();
		if (operands.size() > 0) {
			ModelWithConnections operand = (ModelWithConnections)operands.get(0);
			return operator.getOperatorTypeAsString().toLowerCase() + "\n(" 
					+ indented(getModelWithConnectionText(operand)) + "\n)\n";
		}	
		return "";
	}

	private String getOperatorModelText(OperatorModel operator) {
		if (operator.getOperatorType() == OperatorModel.OPERATOR_NOT) {
			return getOperatorPrefixModelText(operator);
		}
		return getOperatorInfixModelText(operator);
	}

	//	nonFunctionalProperties
	//		dc#description hasValue "The booking ticket is delivered to the buyer via email" 
	//	endNonFunctionalProperties
	private static String getNonFunctionalPropertiesText(AxiomRootModel axiomRoot) {
		String text = "";
		Iterator i = axiomRoot.getNFPs().keySet().iterator();
		while (i.hasNext()) {
			Object key = i.next();
			Object value = axiomRoot.getNFPs().get(key);
			text += key + " hasValue " + Constants.STRING_DEL_START + value + Constants.STRING_DEL_END + "\n"; 
		}
		return "nonFunctionalProperties\n" + indented(text) + "\nendNonFunctionalProperties\n";
	}

	//	?PickupLocation memberOf loc#location and 
	//	?ReturnLocation memberOf loc#location and 
	//	?DoorCount memberOf xsd#integer and   
	//	?MiliageLimited memberOf xsd#boolean and
	private String getAxiomRootModelText(AxiomRootModel axiomRoot) {
		ModelWithConnections top = axiomRoot.getTopModel();
		if (top != null) {
    		return getModelWithConnectionText(top);
		}
		return "";
	}

	private String getMemberOfModelText(MemberOfModel model) {
		String text = "";
		if (model instanceof VariableModel) { 
			text += getVariableModelText((VariableModel)model);
		}
		else if (model instanceof InstanceModel) { 
			text += getInstanceModelText((InstanceModel)model);
		}
		else if (model instanceof ValueInstanceModel) { 
			text += getValueInstanceText((ValueInstanceModel)model);
		}
		// Attach the floating relations which have a parameter connected to this model.
		return text + getChildRelationsText(model.listChildRelations());
	}
	
	private String getModelWithConnectionText(ModelWithConnections model) {
		// Prevent the same model from appearing more than once in the text.
		if (this.visitedModels.contains(model)) {
			return "";
		}
		this.visitedModels.add(model);
		
		String text = "";
		if (model instanceof MemberOfModel) {
			text += getMemberOfModelText((MemberOfModel)model);
		}
		else if (model instanceof RelationModel) { 
			text += getRelationModelText((RelationModel)model);
		}
		else if (model instanceof OperatorModel) { 
			text += getOperatorModelText((OperatorModel)model);
		}
    	return text;
	}

	// exists ?PickupLocation, ?ReturnLocation (
	//		?PickupLocation memberOf loc#location
	//		...
	// )
	private String listVariablesWithExistsFlag(VarNameList varNameList) {
		String text = "";
		Iterator i = varNameList.iterator();
		while (i.hasNext()) {
			VarNameModel varNameModel = (VarNameModel)i.next();
			if (varNameModel.getExistsFlag()) {
				String varName = Constants.VARIABLE_DEL + varNameModel.getName();
				text = (text.length() > 0) ? text + ", " + varName : varName;
			}
		}
		return text;
	}

	// XMLSchema	_"http://www.w3.org/2001/XMLSchema#"
	private String formatNamespaceDeclaration(String prefix, String id) {
		return prefix + "\t_\"" + id + Util.LOCAL_NAME_DELIMITER + "\"";
	}

	// namespace {
	//		XMLSchema	_"http://www.w3.org/2001/XMLSchema#",
	//		example	_"http://www.example.org/ontologies/example#"
	// }
	private String listUsedNamespaces() {
		// Always add the Dublin Core namespace because the axiom uses its tag dc#description.
		String text = formatNamespaceDeclaration("dc", "http://purl.org/dc/elements/1.1");

		Iterator i = this.usedOntologies.iterator();
		while (i.hasNext()) {
			OntologyShell ontology = (OntologyShell)i.next();
			String line = formatNamespaceDeclaration(ontology.getPrefix(), ontology.getOntology().getIdentifier().toString());
			text = (text.length() > 0) ? text + ",\n" + line : line;
		}
		return "namespace {\n" + indented(text) + "\n}\n";
	}

	// The whole axiom text.
    public synchronized String getAxiomText() {
		this.visitedModels.clear();
		this.usedOntologies.clear();

		AxiomRootModel root = this.contentsModel.getAxiomRoot();
    	String definedByText = indented(getAxiomRootModelText(root));

    	String existsVariableList = listVariablesWithExistsFlag(this.contentsModel.getVarNameList());
    	if (existsVariableList.length() > 0) {
    		definedByText = "exists {" + existsVariableList + "} (\n" +
    				definedByText + "\n)";
    	}
    	
    	return listUsedNamespaces() +
    			"axiom " + root.getAxiomName() + "\n" +
				indented(getNonFunctionalPropertiesText(root) +
				"definedBy\n" + definedByText + ".");
    }

}
