package org.infrawebs.ontology.shells;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import org.deri.wsmo4j.io.serializer.wsml.LogExprSerializerWSML;
import org.eclipse.ui.views.properties.IPropertySource;
import org.infrawebs.axiom.plugin.AxiomEditorPlugin;
import org.infrawebs.common.Util;
import org.omwg.logicalexpression.Constants;
import org.omwg.logicalexpression.LogicalExpression;
import org.omwg.ontology.Attribute;
import org.omwg.ontology.Axiom;
import org.omwg.ontology.Concept;
import org.omwg.ontology.Instance;
import org.omwg.ontology.Ontology;
import org.omwg.ontology.OntologyElement;
import org.omwg.ontology.Parameter;
import org.omwg.ontology.Relation;
import org.omwg.ontology.RelationInstance;
import org.omwg.ontology.Value;
import org.omwg.ontology.WsmlDataType;
import org.wsmo.common.Entity;
import org.wsmo.common.IRI;
import org.wsmo.common.Identifier;
import org.wsmo.common.Namespace;
import org.wsmo.common.exception.InvalidModelException;
import org.wsmo.common.exception.SynchronisationException;
import org.wsmo.mediator.Mediator;
import org.wsmostudio.runtime.WSMORuntime;

public class OntologyShell
{
	private Ontology ontology = null;
	private String prefix = null;
	LogExprSerializerWSML leSerializer = null;
    private SelectionPropertySource propsSource;
    
	public OntologyShell(Ontology ontology) {
		this.ontology = ontology;
        this.leSerializer = new LogExprSerializerWSML(ontology);
        this.propsSource = new SelectionPropertySource(this);
	}

	public Ontology getOntology() {
		return this.ontology;
	}

    public IPropertySource getPropertySource() {
        return this.propsSource;
    }

	/**
	 * Returns the local name of the element, possibly prefixed by its namespace.
	 * A prefix is added if the element is defined in an imported ontology.
	 * Locally defined elements have no prefix.
	 */
	public String getQualifiedName(Object o) {
		if (o == null) {
			throw new IllegalArgumentException();
		}
		if (o instanceof Entity) {
			Entity entity = (Entity)o;
			return getQualifiedName(entity.getIdentifier());
		}
		else if (o instanceof WsmlDataType) {
			WsmlDataType type = (WsmlDataType)o;
			return getQualifiedName(type.getIRI());
		}
		else if (o instanceof IRI) {
			IRI iri = (IRI)o;
			String nsPrefix = getNamespacePrefix(iri.toString());
	    	return (nsPrefix != null && nsPrefix.length() > 0)
					? nsPrefix + Constants.SQNAME_DEL + iri.getLocalName()
					: iri.getLocalName();
		}
		else {
			try {
				IRI iri = WSMORuntime.getRuntime().getWsmoFactory().createIRI(o.toString());
				return getQualifiedName(iri);
			}
			catch (IllegalArgumentException ex) {
				return o.toString();
			}
		}
	}
	
	private String getNamespacePrefix(String id) {
		Iterator i = this.ontology.listNamespaces().iterator();
		while (i.hasNext()) {
			Namespace ns = (Namespace)i.next();
			if (id.startsWith(ns.getIRI().toString())) {
				return ns.getPrefix();
			}
		}
		return null;
	}
	
	private String getNFPText(Entity nfpHolder) {
		StringBuffer result = new StringBuffer();
		Map nfps = nfpHolder.listNFPValues();
		Iterator i = nfps.keySet().iterator();
		while (i.hasNext()) {
			// Append the name of the NFP.
			IRI iri = (IRI)i.next();
			result.append("  " + getQualifiedName(iri.toString()) + " = ");
			result.append(getNFPText(nfpHolder, iri) + "\n");
		}
		return (result.length() > 0) ? "NFPs:\n" + result + "\n" : "";
	}
	
	private String getNFPText(Entity nfpHolder, IRI iri) {
		Set values = nfpHolder.listNFPValues(iri);
		if (values.size() > 1) {
			// Append the list of values of the NFP.
			String text = "{";
			Iterator j = values.iterator();
			while (j.hasNext()) {
				Object value = j.next();
				text += value + (j.hasNext() ? ", " : "}");
			}
			return text;
		}
		else if (values.size() > 0) {
			// Append the only value of the NFP.
			Object value = values.iterator().next();
			return value.toString();
		}
		else {
			return "";
		}
	}
	
	public String getElementsAsText(String title, Set set, String suffix) {
		StringBuffer result = new StringBuffer();
		Iterator i = set.iterator();
		while (i.hasNext()) {
			if (result.length() > 0) {
				result.append(", ");
			}
			result.append(getQualifiedName(i.next()));
		}
		return (result.length() > 0) ? title + result + suffix : "";
	}
	
	private String getInstanceAttributesAsText(Instance instance) {
		StringBuffer result = new StringBuffer();
		Map map = instance.listAttributeValues();
		Iterator i = map.keySet().iterator();
		while (i.hasNext()) {
			Attribute attribute = (Attribute)i.next();
			Set values = (Set)map.get(attribute);
			result.append("  ");
			result.append(Util.getLocalName(attribute.getIdentifier()));
			result.append(" = ");
			result.append(getValuesAsText(values));
			result.append("\n");
		}
		return (result.length() > 0) ? "Attributes:\n" + result + "\n" : "";
	}
	
	private String getRelationInstanceParametersAsText(RelationInstance relinstance) {
		StringBuffer result = new StringBuffer();
		int count = relinstance.listParameterValues().size();
		for (byte i = 0; i < count; i++) {
			Parameter param = relinstance.getRelation().getParameter(i);
			result.append(Util.getLocalName(new ParameterShell(param)));
			result.append(" = ");
			try {
				result.append(getQualifiedName(relinstance.getParameterValue(i)));
			}
			catch (InvalidModelException e) {
				// Ignore failure to build text.
			}
			result.append("\n");
		}
		return (result.length() > 0) ? result + "\n" : "";
	}
	
	/**
	 * Returns a comma separated list of the string representations of the values.
	 */
	public String getValuesAsText(Collection values) {
		StringBuffer result = new StringBuffer();
		Iterator i = values.iterator();
		while (i.hasNext()) {
			Value value = (Value)i.next();
			if (result.length() > 0) {
				result.append(", ");
			}
			result.append(getQualifiedName(value));
		}
		return result.toString();
	}

	/**
	 * Used to catch possible exceptions thrown when the concept is not defined.
	 */
	private static Set listAttributesSafe(Concept concept) {
		try {
//			concept = (Concept)AxiomEditorPlugin.ontoStore.findDefinition(concept);
//			if (concept == null) {
//				throw new SynchronisationException(
//						"Cannot use concepts defined in imported ontologies.");
//			}
			return concept.listAttributes();
		}
		catch (SynchronisationException ex) {
			return new HashSet();
		}
	}
	
	public String getProperties(Object item) {
		if (item == null) {
			throw new IllegalArgumentException();
		}
		String text = null;
		if (item instanceof OntologyShell) {
			OntologyShell ao = (OntologyShell)item;
			text = "Ontology: " + ao.getPrefix() + "\n\n" +
					"IRI: " + ao.getOntology().getIdentifier() + "\n\n" +
					"WSML Variant: " + ((ao.getOntology().getWsmlVariant() != null)
					? getQualifiedName(ao.getOntology().getWsmlVariant()) : "unknown");
		}
		else if (item instanceof IRI) {
			IRI iri = (IRI)item;
			text = "Non-functional property: " + getQualifiedName(iri.toString()) + "\n\n" +
					"Values: " + getNFPText(this.ontology, iri);
		}
		else if (item instanceof Namespace) {
			Namespace ns = (Namespace)item;
			text = "Namespace: " + ns.getPrefix() + "\n\n" +
					"IRI: " + ns.getIRI().toString();
		}
		else if (item instanceof Ontology) {
			Ontology onto = (Ontology)item;
			text = "Imported ontology: " + getQualifiedName(onto) + "\n\n" +
					"IRI: " + onto.getIdentifier().toString();
		}
		else if (item instanceof Mediator) {
			Mediator med = (Mediator)item;
			text = "Mediator: " + getQualifiedName(med) + "\n\n" +
					"IRI: " + med.getIdentifier().toString();
		}
		else if (item instanceof Axiom) {
			Axiom axiom = (Axiom)item;
			String definitions = "";
			Iterator i = axiom.listDefinitions().iterator();
			while (i.hasNext()) {
				LogicalExpression expr = (LogicalExpression)i.next();
				definitions += "Definition: " + this.leSerializer.serialize(expr) + "\n\n";
			}
			text = "Axiom: " + getQualifiedName(axiom) + "\n\n" +
					"IRI: " + axiom.getIdentifier().toString() + "\n\n" +
					definitions +
					getNFPText(axiom);
		}
		else if (item instanceof Concept) {
			Concept concept = (Concept)item;
			text = "Concept: " + getQualifiedName(concept) + "\n\n" +
					// Since june 2005, wsmo4j does not support the following:  
					//"Defined by: " + concept.getDefinedBy() + "\n\n" +
					"Attributes: " + listAttributesSafe(concept).size() + "\n\n" +
					getElementsAsText("Super concepts: ", concept.listSuperConcepts(), "\n\n") +
					getNFPText(concept);
		}
		else if (item instanceof AttributeShell) {
			AttributeShell attrInfo = (AttributeShell)item;
			Attribute attribute = attrInfo.getAttribute();
			text = "Attribute: " + Util.getLocalName(attrInfo) + "\n\n" +
					getElementsAsText("Types: ", attribute.listTypes(), "\n\n") +
					"Inheritance status: " + attrInfo.getTypeAsString() + "\n\n" +					
					"Cardinality: " + attrInfo.getCardinalityAsString() + "\n\n" +
					"Constraining: " + attribute.isConstraining() + "\n\n" +
					getNFPText(attribute);
		}
		else if (item instanceof Instance) {
			Instance instance = (Instance)item;
			text = "Instance: " + getQualifiedName(instance) + "\n\n" +
					getInstanceAttributesAsText(instance) +
					getNFPText(instance);
		}
		else if (item instanceof Relation) {
			Relation relation = (Relation)item;
			text = "Relation: " + getQualifiedName(relation) + "\n\n" +
					// Since june 2005, wsmo4j does not support the following:  
					//"Defined by: " + relation.getDefinedBy() + "\n\n" +
					"Parameters: " + relation.listParameters().size() + "\n\n" +
					getElementsAsText("Super relations: ", relation.listSuperRelations(), "\n\n") +
					getNFPText(relation);
		}
		else if (item instanceof ParameterShell) {
			ParameterShell paramInfo = (ParameterShell)item;
			Parameter parameter = paramInfo.getParameter();
			text = "Relation parameter: " + Util.getLocalName(paramInfo) + "\n\n" +
					getElementsAsText("Types: ", parameter.listTypes(), "\n\n") +
					"Constraining: " + parameter.isConstraining();
					// Since june 2005, wsmo4j does not support the following:  
					//getNFPText(parameter);
		}
		else if (item instanceof RelationInstance) {
			RelationInstance relinstance = (RelationInstance)item;
			text = "Relation instance parameters:\n" +
					getRelationInstanceParametersAsText(relinstance) +
					getNFPText(relinstance);
		}
		else {
			text = "N/A";
		}
		return text;
	}

    public void updatePropertiesForSelection(Object selection) {
        this.propsSource.setData(selection);
    }
	// Returns the set of all attributes of the concept and all its superconcepts. 
	public static Set<AttributeShell> listAttributesRecursive(Concept concept) {
		if (concept == null) {
			throw new IllegalArgumentException();
		}
		Map<String, AttributeShell> attributes = new LinkedHashMap<String, AttributeShell>();
		addAttributesRecursive(concept, attributes, concept);
		return new LinkedHashSet<AttributeShell>(attributes.values());
	}
	
	private static void addAttributesRecursive(Concept startConcept,
			Map<String, AttributeShell> attributes, Concept concept) {
		// First, append all attributes of the concept specified.
		Iterator ai = listAttributesSafe(concept).iterator();
		while (ai.hasNext()) {
			Attribute attribute = (Attribute)ai.next();
			String key = Util.getLocalName(attribute.getIdentifier());
			AttributeShell attrInfo = attributes.get(key); 
			if (attrInfo == null) {
				attrInfo = new AttributeShell(attribute, startConcept,
						(concept == startConcept)
						? AttributeShell.NEW : AttributeShell.INHERITED);
				attributes.put(key, attrInfo);
			}
			else {
				attrInfo.setInheritanceType(AttributeShell.OVERRIDEN);
			}
		}
		// Then, append all missing attributes of all superconcepts.
		Iterator ci = concept.listSuperConcepts().iterator();
		while (ci.hasNext()) {
			Concept superConcept = (Concept)ci.next();

			// The superConcept may be defined in another ontology.
			// If the superConcept is defined in another ontology, let's find it and use its definition
			Concept superConceptDefinition = (Concept)AxiomEditorPlugin.ontoStore.findDefinition(superConcept);
//			if (superConcept != superConceptDefinition)
//			{
//				// we have a concept which is defined in another ontology!
//			}
			
			if (superConceptDefinition != null) {
				addAttributesRecursive(startConcept, attributes, superConcept);
			}
		}
	}

	/**
	 * @return The first ontology element with the supplied identifier.
	 * @note It is possible that several kinds of elements have the same identifier.
	 */
	private OntologyElement findFirstOntologyElement(Identifier id, Class elementClass) {
		Iterator i = this.ontology.findEntity(id).iterator();
		while (i.hasNext()) {
			Object entity = i.next();
			if (elementClass.isInstance(entity)) {
				return (OntologyElement)entity;
			}
		}
		return null;
	}

	/**
	 * Returns the object which contains the full definition of the
	 * identifier of the element passed.
	 * Returns null if the identifier is defined in another ontology
	 * or is just referenced from this ontology.
	 */
	public OntologyElement findDefinition(OntologyElement element) {
		if (element == null) {
			throw new IllegalArgumentException();
		}
		// If an attribute/parameter's concept/relation is defined in this intology,
		// then the attribute/parameter is also considered defined in it.
		if (element instanceof RefinableShell) {
			OntologyElement parent = findDefinition(((RefinableShell)element).getParent());
			return (parent != null) ? element : null;
		}
		return findFirstOntologyElement(element.getIdentifier(), element.getClass());
	}

	/**
	 * Works for entities only. Doesn't work for attributes/parameters.
	 */
	public OntologyElement findDefinition(Identifier id, Class elementClass) {
		if (id == null) {
			throw new IllegalArgumentException();
		}
		return findFirstOntologyElement(id, elementClass);
	}

	public String getPrefix() {
		return this.prefix;
	}

	public void setPrefix(String prefix) {
		this.prefix = prefix;
	}

}
