package org.infrawebs.ontology.shells;

import java.util.*;

import org.eclipse.ui.views.properties.*;
import org.infrawebs.common.Util;
import org.omwg.logicalexpression.LogicalExpression;
import org.omwg.ontology.*;
import org.wsmo.common.*;
import org.wsmo.common.exception.InvalidModelException;
import org.wsmo.mediator.Mediator;

public class SelectionPropertySource implements IPropertySource {

    private OntologyShell ontoShell;
    private Object selection = null;
    
    private Map<Integer, String> id2value = null;
    
    private List<IPropertyDescriptor> props;
    private int idCounter = 0;
    
    SelectionPropertySource(OntologyShell oShell) {
        this.ontoShell = oShell;
        initData();
    }
    
    public void setData(Object target) {
        if (target != null 
                && target.equals(this.selection)) {
            return;
        }
        this.selection = target;
        initData();
    }
    
    private void initData() {
    	this.idCounter = 0;
        if (this.id2value == null) {
        	this.props = new LinkedList<IPropertyDescriptor>();
        	this.id2value = new HashMap<Integer, String>();
        }
        else {
        	this.props.clear();
        	this.id2value.clear();
        }
        if (this.selection == null) {
            return;
        }

        if (this.selection instanceof OntologyShell) {
            OntologyShell ao = (OntologyShell)this.selection;
            this.addPropValue(null, "Ontology:", ao.getPrefix());
            this.addPropValue(null, "IRI:", ao.getOntology().getIdentifier().toString());
            this.addPropValue(null, "WSML Variant:", (ao.getOntology().getWsmlVariant() != null)
                    ? ao.getQualifiedName(ao.getOntology().getWsmlVariant()) : "unknown");
        }
        else if (this.selection instanceof IRI) {
            IRI iri = (IRI)this.selection;
            this.addPropValue(null, "Non-functional property", this.ontoShell.getQualifiedName(iri.toString()));
            this.addPropValue(null, "Value(s)", this.ontoShell.getOntology().listNFPValues(iri).toString());
        }
        else if (this.selection instanceof Namespace) {
            Namespace ns = (Namespace)this.selection;
            this.addPropValue(null, "Namespace:", ns.getPrefix());
            this.addPropValue(null, "IRI:", ns.getIRI().toString());
        }
        else if (this.selection instanceof Ontology) {
            Ontology onto = (Ontology)this.selection;
            this.addPropValue(null, "Imported ontology:", this.ontoShell.getQualifiedName(onto));
            this.addPropValue(null, "IRI:", onto.getIdentifier().toString());
        }
        else if (this.selection instanceof Mediator) {
            Mediator med = (Mediator)this.selection;
            this.addPropValue(null, "Mediator:", this.ontoShell.getQualifiedName(med));
            this.addPropValue(null, "IRI:", med.getIdentifier().toString());
        }
        else if (this.selection instanceof Axiom) {
            Axiom axiom = (Axiom)this.selection;
            this.addPropValue(null, "Axiom:", this.ontoShell.getQualifiedName(axiom));
            this.addPropValue(null, "IRI:", axiom.getIdentifier().toString());
            Iterator i = axiom.listDefinitions().iterator();
            while (i.hasNext()) {
                LogicalExpression expr = (LogicalExpression)i.next();
                this.addPropValue(null, "Definition:", this.ontoShell.leSerializer.serialize(expr));
            }
            processNFPs(axiom);
        }
        else if (this.selection instanceof Concept) {
            Concept concept = (Concept)this.selection;
            this.addPropValue(null, "Concept:", this.ontoShell.getQualifiedName(concept));
            this.addPropValue(null, "Attributes:", ""+concept.listAttributes().size());
            processElementsAsText("Super concepts: ", concept.listSuperConcepts());
            processNFPs(concept);
        }
        else if (this.selection instanceof AttributeShell) {
            AttributeShell attrInfo = (AttributeShell)this.selection;
            Attribute attribute = attrInfo.getAttribute();
            this.addPropValue(null, "Attribute:", Util.getLocalName(attrInfo));
            processElementsAsText("Types: ", attribute.listTypes());
            this.addPropValue(null, "Inheritance status: ", attrInfo.getTypeAsString());                  
            this.addPropValue(null, "Cardinality:", attrInfo.getCardinalityAsString());
            this.addPropValue(null, "Constraining:", "" + attribute.isConstraining());
            processNFPs(attribute);
        }
        else if (this.selection instanceof Instance) {
            Instance instance = (Instance)this.selection;
            this.addPropValue(null, "Instance:", this.ontoShell.getQualifiedName(instance));
            processInstanceAttributesAsText(instance);
            processNFPs(instance);
        }
        else if (this.selection instanceof Relation) {
            Relation relation = (Relation)this.selection;
            addPropValue(null, "Relation:", this.ontoShell.getQualifiedName(relation));
            addPropValue(null, "Parameters:", "" + relation.listParameters().size());
            processElementsAsText("Super relations:", relation.listSuperRelations());
            processNFPs(relation);
        }
        else if (this.selection instanceof ParameterShell) {
            ParameterShell paramInfo = (ParameterShell)this.selection;
            Parameter parameter = paramInfo.getParameter();
            addPropValue(null, "Relation parameter:", Util.getLocalName(paramInfo));
            processElementsAsText("Types:", parameter.listTypes());
            addPropValue(null, "Constraining: ", "" + parameter.isConstraining());
        }
        else if (this.selection instanceof RelationInstance) {
            RelationInstance relinstance = (RelationInstance)this.selection;
            processRelationInstanceParametersAsText(relinstance);
            processNFPs(relinstance);
        }
    }
    
    private void processNFPs(Entity nfpHolder) {
        Map nfps = nfpHolder.listNFPValues();
        Iterator i = nfps.keySet().iterator();
        while (i.hasNext()) {
            IRI iri = (IRI)i.next();
            addPropValue("NFP", 
            		this.ontoShell.getQualifiedName(iri.toString()), 
                         nfps.get(iri).toString());
        }
    }
    
    private void processRelationInstanceParametersAsText(RelationInstance relinstance) {

        int count = relinstance.listParameterValues().size();
        for (byte i = 0; i < count; i++) {
            Parameter param = relinstance.getRelation().getParameter(i);
            String paramVal = null;
            try {
                paramVal = this.ontoShell.getQualifiedName(relinstance.getParameterValue(i));
            }
            catch (InvalidModelException e) {
                paramVal = "<error retrieving value>";
            }
            addPropValue("Relation instance parameters", 
                         Util.getLocalName(new ParameterShell(param)),
                         paramVal);
        }
    }

    private void processElementsAsText(String title, Set set) {
        StringBuffer result = new StringBuffer();
        Iterator i = set.iterator();
        while (i.hasNext()) {
            if (result.length() > 0) {
                result.append(", ");
            }
            result.append(this.ontoShell.getQualifiedName(i.next()));
        }
        if (result.length() > 0) {
            addPropValue(null, title, result.toString());
        }
	}

    private void processInstanceAttributesAsText(Instance instance) {
    	Map map = instance.listAttributeValues();
    	Iterator i = map.keySet().iterator();
    	while (i.hasNext()) {
        	IRI iri = (IRI)i.next();
            Set values = (Set)map.get(iri);
            addPropValue("Attributes", Util.getLocalName(iri),
            		this.ontoShell.getValuesAsText(values));
    	}
    }
    
    private void addPropValue(String cat, String propName, String value) {
        if (cat == null) {
            cat = "Common";
        }
        Integer id = new Integer(this.idCounter++);
        PropertyDescriptor idDesc = new PropertyDescriptor(id, propName);
        idDesc.setCategory(cat);

        this.props.add(idDesc);
        this.id2value.put(id, value);
    }
    
    public Object getEditableValue() {
        return null;
    }

    public IPropertyDescriptor[] getPropertyDescriptors() {
        return this.props.toArray(new IPropertyDescriptor[this.props.size()]);
    }

    public Object getPropertyValue(Object id) {
        return this.id2value.get(id);
    }

    public boolean isPropertySet(Object id) {
        return false;
    }

    public void resetPropertyValue(Object id) {
    	// This property source is read-only.
    }

    public void setPropertyValue(Object id, Object value) {
    	// This property source is read-only.
    }

}
