package org.infrawebs.ontology.ui;

import java.util.*;
import java.util.List;

import org.eclipse.gef.dnd.TemplateTransfer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.*;
import org.eclipse.swt.events.*;
import org.eclipse.swt.layout.*;
import org.eclipse.swt.widgets.*;
import org.infrawebs.common.Util;
import org.infrawebs.ontology.OntoFilter;
import org.infrawebs.ontology.shells.*;
import org.omwg.ontology.*;
import org.wsmo.common.*;
import org.wsmostudio.ui.views.navigator.WSMONavigator;

public class OntoPane extends Composite {
	
	private Tree tree = null;
	private TreeItem rootMetadataNode = null;
	private TreeItem rootConceptNode = null;
	private TreeItem rootRelationNode = null;
	
	OntologyShell ontology = null;
	private OntoFilter ontoFilter = null;

	public OntoPane(Composite parent) {
		super(parent, SWT.NONE);
		setLayout(new GridLayout(1, false));
        createTree();
	}

	private class TreeSelectionListener extends SelectionAdapter
	{
		@Override
		public void widgetSelected(SelectionEvent e) {
			OntoPane.this.ontology.updatePropertiesForSelection(getSelection());
            WSMONavigator.propsViewRefresh();
 		}
	}

	private void createTree() {
		this.tree = new Tree(this, SWT.BORDER);
        this.tree.setLayoutData(new GridData(GridData.FILL_BOTH));
		this.tree.addSelectionListener(new TreeSelectionListener());

		// Create the drag source on the tree
		DragSource ds = new DragSource(this.tree, DND.DROP_MOVE);
		ds.setTransfer(new Transfer[] {TemplateTransfer.getInstance()});
		ds.addDragListener(new DragSourceAdapter() {
			@Override
			public void dragSetData(DragSourceEvent event) {
				// Next line was required by Eclipse 3.0 but not 3.1.
				//TemplateTransfer.getInstance().setObject(getSelection());
				event.data = getSelection();
			}
		});
	}

	public OntologyShell getOntology() {
		return this.ontology;
	}

	public void setOntology(OntologyShell ontology, OntoFilter ontoFilter) {
		this.ontology = ontology;
		this.ontoFilter = ontoFilter;
		buildOntologyTree();
	}

	public Object getSelection() {
		return this.tree.getSelection()[0].getData();
	}

    private TreeItem addNode(TreeItem parent, String text, Object data, int imageIndex) {
        TreeItem node = new TreeItem(parent, SWT.NONE);
        node.setText(text);
        node.setData(data);
		node.setImage(OntoImages.getImage(imageIndex));
        return node;
	}

	// Appends to the 'found' list all tree items which point to the specified ontology element.
	// The search starts at the 'rootNode' item and continues down into its subtree.
	private void findTreeItems(TreeItem rootNode, OntologyElement element, List<TreeItem> found) {
		// Compare elements using the equals() method (not by reference).
		if (rootNode != null && rootNode.getData() != null &&
				rootNode.getData().equals(element)) {
			found.add(rootNode);
		}
		// Search recursively.
		TreeItem[] children = (rootNode != null) ? rootNode.getItems() : this.tree.getItems();
		for (int i = 0; i < children.length; i++) {
			findTreeItems(children[i], element, found);
		}
	}

	private void buildOntologyTree() {
		this.tree.removeAll();

		if (this.ontoFilter.showMetadata) {
			addAllMetadata();
		}

		if (this.ontoFilter.showConcepts) {
			addAllConcepts();
			Util.ExpandFullyTreeItem(this.rootConceptNode, true);
		}
		else if (this.ontoFilter.showInstances) {
			addAllInstances();
			Util.ExpandFullyTreeItem(this.rootConceptNode, true);
		}

		if (this.ontoFilter.showRelations) {
			addAllRelations();
			Util.ExpandFullyTreeItem(this.rootRelationNode, true);
		}
	}

	private void addAllMetadata() {
		this.rootMetadataNode = new TreeItem(this.tree, SWT.NONE);
		this.rootMetadataNode.setText("More information");
		this.rootMetadataNode.setImage(OntoImages.getImage(OntoImages.SPHERE));
		this.rootMetadataNode.setData(this.ontology);
		
		// Add non-functional properties.
		TreeItem rootNfpNode = addNode(this.rootMetadataNode, "Non-functional properties",
				this.ontology, OntoImages.NFP);
		Iterator nfpi = this.ontology.getOntology().listNFPValues().keySet().iterator();
		while (nfpi.hasNext()) {
			IRI iri = (IRI)nfpi.next();
			addNode(rootNfpNode, this.ontology.getQualifiedName(iri.toString()),
					iri, OntoImages.NFP);
		}
		Util.ExpandFullyTreeItem(rootNfpNode, true);

		// Add namespaces.
		TreeItem rootNamespaceNode = addNode(this.rootMetadataNode, "Namespaces",
				this.ontology, OntoImages.NAMESPACE);
		Iterator nsi = this.ontology.getOntology().listNamespaces().iterator();
		while (nsi.hasNext()) {
			Namespace ns = (Namespace)nsi.next();
			addNode(rootNamespaceNode, ns.getPrefix(),
					ns, OntoImages.NAMESPACE);
		}
		Util.ExpandFullyTreeItem(rootNamespaceNode, true);

		// Add imported ontologies.
		TreeItem rootOntologyNode = addNode(this.rootMetadataNode, "Imported ontologies",
				this.ontology, OntoImages.IMPORTED_ONTOLOGY);
		Iterator ontoi = this.ontology.getOntology().listOntologies().iterator();
		while (ontoi.hasNext()) {
			Ontology onto = (Ontology)ontoi.next();
			addNode(rootOntologyNode, Util.getLocalName(onto),
					onto, OntoImages.IMPORTED_ONTOLOGY);
		}
		Util.ExpandFullyTreeItem(rootOntologyNode, true);

		// Add used mediators.
		TreeItem rootMediatorNode = addNode(this.rootMetadataNode, "Used mediators",
				this.ontology, OntoImages.MEDIATOR);
		Iterator medi = this.ontology.getOntology().listMediators().iterator();
		while (medi.hasNext()) {
			IRI medID = (IRI)medi.next();
			addNode(rootMediatorNode, this.ontology.getQualifiedName(medID),
					medID, OntoImages.MEDIATOR);
		}
		Util.ExpandFullyTreeItem(rootMediatorNode, true);

		// Add defined axioms.
		TreeItem rootAxiomNode = addNode(this.rootMetadataNode, "Defined axioms",
				this.ontology, OntoImages.AXIOM);
		Iterator axi = this.ontology.getOntology().listAxioms().iterator();
		while (axi.hasNext()) {
			Axiom axiom = (Axiom)axi.next();
			addNode(rootAxiomNode, this.ontology.getQualifiedName(axiom),
					axiom, OntoImages.AXIOM);
		}
		Util.ExpandFullyTreeItem(rootAxiomNode, true);
	}
	
	private void addAllConcepts() {
		this.rootConceptNode = new TreeItem(this.tree, SWT.NONE);
		this.rootConceptNode.setText("Concepts");
		this.rootConceptNode.setImage(OntoImages.getImage(OntoImages.CONCEPT_DISABLED));
		
		List<TreeItem> conceptNodes = new ArrayList<TreeItem>(); 
		Iterator ci = this.ontology.getOntology().listConcepts().iterator();

		while (ci.hasNext()) {
			Concept concept = (Concept)ci.next();
			// Skip incompatible concepts.
			if (!this.ontoFilter.isCompatible(concept)) {
				continue;
			}
			
			conceptNodes.clear();
			addConceptNodes(concept, conceptNodes);
			for (TreeItem conceptNode: conceptNodes) { 
				if (this.ontoFilter.showAttributes) {
					attachAttributeNodes(conceptNode);
				}
				if (this.ontoFilter.showInstances) {
					attachInstanceNodes(conceptNode);
				}
			}
		}
	}


	public void updateAllConcepts() {
		
		//updateConceptRecursively(rootConceptNode);
		//addAllConcepts();

		
		this.rootConceptNode.removeAll();
		
		List<TreeItem> conceptNodes = new ArrayList<TreeItem>(); 
		Iterator ci = this.ontology.getOntology().listConcepts().iterator();

		while (ci.hasNext()) {
			Concept concept = (Concept)ci.next();
			// Skip incompatible concepts.
			if (!this.ontoFilter.isCompatible(concept)) {
				continue;
			}
			
			conceptNodes.clear();
			addConceptNodes(concept, conceptNodes);
			for (TreeItem conceptNode: conceptNodes) { 
				if (this.ontoFilter.showAttributes) {
					attachAttributeNodes(conceptNode);
				}
				if (this.ontoFilter.showInstances) {
					attachInstanceNodes(conceptNode);
				}
			}
		}
		
		Util.ExpandFullyTreeItem(this.rootConceptNode, true);		
	}
	
	private void addAllInstances() {
		this.rootConceptNode = new TreeItem(this.tree, SWT.NONE);
		this.rootConceptNode.setText("Instances");
		this.rootConceptNode.setImage(OntoImages.getImage(OntoImages.INSTANCE_DISABLED));
		
		List<TreeItem> conceptNodes = new ArrayList<TreeItem>(); 
		Iterator ii = this.ontology.getOntology().listInstances().iterator();
		while (ii.hasNext()) {
			Instance instance = (Instance)ii.next();
			// Skip incompatible instances.
			if (!this.ontoFilter.isCompatible(instance)) {
				continue;
			}

			// First appends all concepts this instance belongs to.
			// Then attaches the instance to the concept nodes.
			Iterator ci = instance.listConcepts().iterator();
			while (ci.hasNext()) {
				Concept concept = (Concept)ci.next();
				conceptNodes.clear();
				addConceptNodes(concept, conceptNodes);
				for (TreeItem conceptNode: conceptNodes) { 
					if (this.ontoFilter.showAttributes) {
						attachAttributeNodes(conceptNode);
					}
					addNode(conceptNode, this.ontology.getQualifiedName(instance),
							instance, OntoImages.INSTANCE);
				}
			}
		}
	}
	
	// Recursively adds to the tree a concept and its superconcepts.
	// Appends to conceptNodes all nodes representing the concept. Since a
	// concept may have many superconcepts, several nodes at different
	// places in the tree may represent the same concept.
	private void addConceptNodes(Concept concept, List<TreeItem> conceptNodes) {
		// Avoid concept duplication. If any nodes exist for that concept, just append 
		// all of them to the list and return.
		int oldSize = conceptNodes.size();
		findTreeItems(this.rootConceptNode, concept, conceptNodes);
		if (conceptNodes.size() > oldSize) {
			return;
		}
		
		// Add all superconcepts before adding this concept.
		List<TreeItem> superConceptNodes = new ArrayList<TreeItem>(); 
		Iterator i = concept.listSuperConcepts().iterator();
		if (i.hasNext()) {
			// Not a root concept.
			while (i.hasNext()) {
				Concept superConcept = (Concept)i.next();
				addConceptNodes(superConcept, superConceptNodes);
			}
		}
		else {
			// This is a root concept (bottom of recursion).
			superConceptNodes.add(this.rootConceptNode);
		}

		// Finally, append this concept as a child to all superconcepts found.
		i = superConceptNodes.iterator();
		while (i.hasNext()) {
			int imageIndex = this.ontoFilter.isCompatible(concept) && this.ontoFilter.selectConcepts
					? OntoImages.CONCEPT
					: OntoImages.CONCEPT_DISABLED;
	    	TreeItem conceptNode = addNode((TreeItem)i.next(),
	    			this.ontology.getQualifiedName(concept), concept, imageIndex);
	    	conceptNodes.add(conceptNode);
		}
	}
	
	private void attachAttributeNodes(TreeItem conceptNode) {
		Concept concept = (Concept)conceptNode.getData();
	    Iterator ai = OntologyShell.listAttributesRecursive(concept).iterator();
	    while (ai.hasNext()) {
	    	AttributeShell attrInfo = (AttributeShell)ai.next();
	    	addNode(conceptNode, Util.getLocalName(attrInfo),
	    			attrInfo, attrInfo.getInheritanceType());
	    }
	}

	
	private void attachInstanceNodes(TreeItem conceptNode) {
		Concept concept = (Concept)conceptNode.getData();
		Iterator ii = concept.listInstances().iterator();
		while (ii.hasNext()) {
			Instance instance = (Instance)ii.next();
			int imageIndex = this.ontoFilter.isCompatible(instance) && this.ontoFilter.selectInstances
					? OntoImages.INSTANCE
					: OntoImages.INSTANCE_DISABLED;
			addNode(conceptNode, this.ontology.getQualifiedName(instance), instance, imageIndex);
		}
	}

	private void addAllRelations() {
		this.rootRelationNode = new TreeItem(this.tree, SWT.NONE);
		this.rootRelationNode.setText("Relations");
		this.rootRelationNode.setImage(OntoImages.getImage(OntoImages.RELATION_DISABLED));
		
		List<TreeItem> relationNodes = new ArrayList<TreeItem>(); 
        Iterator ri = this.ontology.getOntology().listRelations().iterator();
        while (ri.hasNext()) {
        	Relation relation = (Relation)ri.next();
			// Skip incompatible relations.
			if (!this.ontoFilter.isCompatible(relation)) {
				continue;
			}
			
        	relationNodes.clear();
        	addRelationNodes(relation, relationNodes);
        	for (TreeItem relationNode: relationNodes) { 
       			attachRelationParameterNodes(relationNode);
        		attachRelationInstanceNodes(relationNode);
        	}
        }
	}
	
	// Recursively adds to the tree a relation and its superrelations.
	// Appends to the list all tree items that point to the specified relation. 
	private void addRelationNodes(Relation relation, List<TreeItem> relationNodes) {
		// Avoid relation duplication. If any nodes exist for that relation, just append 
		// all of them to the list and return.
		int oldSize = relationNodes.size();
		findTreeItems(this.rootRelationNode, relation, relationNodes);
		if (relationNodes.size() > oldSize) {
			return;
		}
		
		// Add all superrelations before adding this relation.
		List<TreeItem> superRelationNodes = new ArrayList<TreeItem>(); 
		Iterator i = relation.listSuperRelations().iterator();
		if (i.hasNext()) {
			// Not a root relation.
			while (i.hasNext()) {
				Relation superRelation = (Relation)i.next();
				addRelationNodes(superRelation, superRelationNodes);
			}
		}
		else {
			// This is a root relation (bottom of recursion).
			superRelationNodes.add(this.rootRelationNode);
		}

		// Finally, append this relation as a child to all superrelations found.
		i = superRelationNodes.iterator();
		while (i.hasNext()) {
			int imageIndex = this.ontoFilter.isCompatible(relation) && this.ontoFilter.selectRelations
					? OntoImages.RELATION
					: OntoImages.RELATION_DISABLED;
	    	TreeItem relationNode = addNode((TreeItem)i.next(),
	    			this.ontology.getQualifiedName(relation), relation, imageIndex);
	    	relationNodes.add(relationNode);
		}
	}
	
	private void attachRelationParameterNodes(TreeItem relationNode) {
		Relation relation = (Relation)relationNode.getData();
		int count = relation.listParameters().size();
	    for (byte i = 0; i < count; i++) {
	    	Parameter param = relation.getParameter(i);
			int imageIndex = this.ontoFilter.isCompatible(param) && this.ontoFilter.selectParameters
					? OntoImages.RELATION_PARAMETER
					: OntoImages.RELATION_PARAMETER_DISABLED;
			ParameterShell paramInfo = new ParameterShell(param);
	    	addNode(relationNode, Util.getLocalName(paramInfo), paramInfo, imageIndex);
	    }
	}

	private void attachRelationInstanceNodes(TreeItem relationNode) {
		Relation relation = (Relation)relationNode.getData();
		Iterator ii = relation.listRelationInstances().iterator();
		while (ii.hasNext()) {
			RelationInstance relinstance = (RelationInstance)ii.next();
			List values = relinstance.listParameterValues();
			String nodeText = getOntology().getValuesAsText(values);
			addNode(relationNode, "(" + nodeText + ")",
					relinstance, OntoImages.RELATION_INSTANCE);
		}
	}

	// Selects the first tree item which contains the same ontology element.
	public boolean jumpToElement(OntologyElement element) {
		List<TreeItem> foundNodes = new ArrayList<TreeItem>();
		if ((element instanceof Relation) 
				|| (element instanceof RelationInstance)
				|| (element instanceof ParameterShell)) {
			findTreeItems(this.rootRelationNode, element, foundNodes);
		}
		else {
			findTreeItems(this.rootConceptNode, element, foundNodes);
		}

		if (foundNodes.size() > 0) {
			this.tree.setSelection(new TreeItem[] {foundNodes.get(0)});
			return true;
		} 
		return false;
	}

	public boolean isEmpty() {
		return (this.rootConceptNode == null || this.rootConceptNode.getItemCount() == 0) &&
				(this.rootRelationNode == null || this.rootRelationNode.getItemCount() == 0);
	}

	public OntoFilter getOntoFilter() {
		return this.ontoFilter;
	}

	public Tree getTree() {
		return this.tree;
	}

}
